import inspect

import tensorflow as tf
from absl.testing import parameterized
from tensorflow import keras

from keras_aug import layers
from keras_aug.datapoints import bounding_box
from keras_aug.layers import augmentation
from keras_aug.layers import preprocessing
from keras_aug.utils.augmentation import BOUNDING_BOXES
from keras_aug.utils.augmentation import IMAGES
from keras_aug.utils.augmentation import LABELS

#   (
#       name,
#       layer_cls,
#       args,
#       is_bbox_compatible,
#   )
# all configurations should be expanded for readability
GENERAL_TESTS = [
    (
        "AugMix",
        layers.AugMix,
        {"value_range": (0, 255)},
        False,
    ),
    (
        "RandAugment",
        layers.RandAugment,
        {"value_range": (0, 255)},
        True,
    ),
    (
        "TrivialAugmentWide",
        layers.TrivialAugmentWide,
        {"value_range": (0, 255)},
        True,
    ),
    (
        "RandomAffine",
        layers.RandomAffine,
        {
            "rotation_factor": 10,
            "translation_height_factor": 0.1,
            "translation_width_factor": 0.1,
            "zoom_height_factor": 0.1,
            "zoom_width_factor": 0.1,
            "shear_height_factor": 0.1,
            "shear_width_factor": 0.1,
        },
        True,
    ),
    (
        "RandomCrop",
        layers.RandomCrop,
        {"height": 2, "width": 2},
        True,
    ),
    (
        "RandomCropAndResize",
        layers.RandomCropAndResize,
        {
            "height": 2,
            "width": 2,
            "crop_area_factor": (0.8, 1.0),
            "aspect_ratio_factor": (3 / 4, 4 / 3),
        },
        True,
    ),
    (
        "RandomFlip",
        layers.RandomFlip,
        {"mode": "horizontal"},
        True,
    ),
    (
        "RandomResize",
        layers.RandomResize,
        {"heights": [2]},
        True,
    ),
    (
        "RandomRotate",
        layers.RandomRotate,
        {"factor": 10},
        True,
    ),
    (
        "RandomZoomAndCrop",
        layers.RandomZoomAndCrop,
        {"height": 2, "width": 2, "scale_factor": (0.8, 1.25)},
        True,
    ),
    (
        "ChannelShuffle",
        layers.ChannelShuffle,
        {"groups": 3},
        True,
    ),
    (
        "RandomBlur",
        layers.RandomBlur,
        {"factor": (3, 7)},
        True,
    ),
    (
        "RandomChannelShift",
        layers.RandomChannelShift,
        {"value_range": (0, 255), "factor": 0.1},
        True,
    ),
    (
        "RandomCLAHE",
        layers.RandomCLAHE,
        {"value_range": (0, 255), "factor": (2, 10), "tile_grid_size": (4, 4)},
        True,
    ),
    (
        "RandomColorJitter",
        layers.RandomColorJitter,
        {
            "value_range": (0, 255),
            "brightness_factor": 0.1,
            "contrast_factor": 0.1,
            "saturation_factor": 0.1,
            "hue_factor": 0.1,
        },
        True,
    ),
    (
        "RandomGamma",
        layers.RandomGamma,
        {"value_range": (0, 255), "factor": 0.1},
        True,
    ),
    (
        "RandomGaussianBlur",
        layers.RandomGaussianBlur,
        {"kernel_size": 3, "factor": 2.0},
        True,
    ),
    (
        "RandomHSV",
        layers.RandomHSV,
        {
            "value_range": (0, 255),
            "hue_factor": 0.1,
            "saturation_factor": 0.1,
            "value_factor": 0.1,
        },
        True,
    ),
    (
        "RandomJpegQuality",
        layers.RandomJpegQuality,
        {
            "value_range": (0, 255),
            "factor": (75, 100),
        },
        True,
    ),
    (
        "RandomPosterize",
        layers.RandomPosterize,
        {"value_range": (0, 255), "factor": (5, 8)},
        True,
    ),
    (
        "RandomSharpness",
        layers.RandomSharpness,
        {"value_range": (0, 255), "factor": 0.1},
        True,
    ),
    (
        "RandomSolarize",
        layers.RandomSolarize,
        {
            "value_range": (0, 255),
            "threshold_factor": 10,
            "addition_factor": 10,
        },
        True,
    ),
    (
        "CutMix",
        layers.CutMix,
        {"alpha": 1.0},
        False,
    ),
    (
        "MixUp",
        layers.MixUp,
        {},
        True,
    ),
    (
        "Mosaic",
        layers.Mosaic,
        {
            "height": 100,
            "width": 100,
        },
        True,
    ),
    (
        "RandomChannelDropout",
        layers.RandomChannelDropout,
        {},
        True,
    ),
    (
        "RandomCutout",
        layers.RandomCutout,
        {"height_factor": 0.3, "width_factor": 0.3},
        True,
    ),
    (
        "RandomErase",
        layers.RandomErase,
        {"area_factor": (0.02, 0.4), "aspect_ratio_factor": (0.3, 1.0 / 0.3)},
        False,
    ),
    (
        "RandomGridMask",
        layers.RandomGridMask,
        {
            "size_factor": (0.5, 1.0),
            "ratio_factor": (0.6, 0.6),
            "rotation_factor": (-10, 10),
        },
        True,
    ),
    (
        "CenterCrop",
        layers.CenterCrop,
        {"height": 2, "width": 2},
        True,
    ),
    (
        "PadIfNeeded",
        layers.PadIfNeeded,
        {"min_height": 2, "min_width": 2},
        True,
    ),
    (
        "Resize",
        layers.Resize,
        {"height": 2, "width": 2},
        True,
    ),
    (
        "AutoContrast",
        layers.AutoContrast,
        {"value_range": (0, 255)},
        True,
    ),
    (
        "Equalize",
        layers.Equalize,
        {"value_range": (0, 255)},
        True,
    ),
    (
        "Grayscale",
        layers.Grayscale,
        {"output_channels": 3},
        True,
    ),
    (
        "Identity",
        layers.Identity,
        {},
        True,
    ),
    (
        "Invert",
        layers.Invert,
        {"value_range": (0, 255)},
        True,
    ),
    (
        "Normalize",
        layers.Normalize,
        {"value_range": (0, 255)},
        True,
    ),
    (
        "Rescale",
        layers.Rescale,
        {"scale": 1.0 / 255.0},
        True,
    ),
    (
        "SanitizeBoundingBox",
        layers.SanitizeBoundingBox,
        {"min_size": 10},
        True,
    ),
]

BUILD_IN_RUNTIME = [
    (
        "RandomApply",
        layers.RandomApply,
        {"name": "layer", "value": "single", "args": {}},
        layers.RandomChannelDropout,
        {},
    ),
    (
        "RandomApplyBatchwise",
        layers.RandomApply,
        {"name": "layer", "value": "single", "args": {"batchwise": True}},
        layers.RandomChannelDropout,
        {},
    ),
    (
        "RandomChoice",
        layers.RandomChoice,
        {"name": "layers", "value": "multiple", "args": {}},
        layers.RandomChannelDropout,
        {},
    ),
    (
        "RandomChoiceBatchwise",
        layers.RandomChoice,
        {"name": "layers", "value": "multiple", "args": {"batchwise": True}},
        layers.RandomChannelDropout,
        {},
    ),
    (
        "RepeatedAugment",
        layers.RepeatedAugment,
        {"name": "layers", "value": "multiple", "args": {}},
        layers.RandomColorJitter,
        {"value_range": (0, 255), "brightness_factor": (1.5, 1.5)},
    ),
]


class MixedPrecisionFloat16Test(tf.test.TestCase, parameterized.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        tf.debugging.enable_check_numerics()
        keras.mixed_precision.set_global_policy("mixed_float16")

    def test_all_2d_aug_layers_included(self):
        base_cls = layers.VectorizedBaseRandomLayer
        cls_spaces = [augmentation, preprocessing]
        all_2d_aug_layers = []
        for cls_space in cls_spaces:
            all_2d_aug_layers.extend(
                inspect.getmembers(cls_space, predicate=inspect.isclass)
            )
        all_2d_aug_layer_names = set(
            item[0]
            for item in all_2d_aug_layers
            if issubclass(item[1], base_cls)
        )

        general_names = set(item[0] for item in GENERAL_TESTS)
        build_names = set(item[0] for item in BUILD_IN_RUNTIME)
        all_test_names = general_names.union(build_names)

        for name in all_2d_aug_layer_names:
            self.assertIn(name, all_test_names, msg=f"{name} not found")

    @parameterized.named_parameters(*GENERAL_TESTS)
    def test_run_in_mixed_precision(self, layer_cls, args, is_bbox_compatible):
        images = tf.cast(
            tf.random.uniform(shape=(2, 224, 224, 3)) * 255.0,
            dtype=tf.float64,
        )
        labels = tf.cast(
            tf.random.uniform(shape=(2, 1)) * 10.0, dtype=tf.float64
        )
        bounding_boxes = {
            "boxes": tf.ragged.constant(
                [
                    [[10, 10, 20, 20], [100, 100, 150, 150]],
                    [[200, 200, 400, 400]],
                ],
                dtype=tf.float32,
            ),
            "classes": tf.ragged.constant([[0, 1], [2]], dtype=tf.float32),
        }
        if is_bbox_compatible:
            try:
                layer = layer_cls(**args, bounding_box_format="xyxy")
            except TypeError:
                layer = layer_cls(**args)
            inputs = {
                IMAGES: images,
                LABELS: labels,
                BOUNDING_BOXES: bounding_boxes,
            }
        else:
            layer = layer_cls(**args)
            inputs = {IMAGES: images, LABELS: labels}

        outputs = layer(inputs)

        if is_bbox_compatible:
            self.assertDTypeEqual(outputs[IMAGES], tf.float16)
            dense_bounding_boxes = bounding_box.to_dense(
                outputs[BOUNDING_BOXES]
            )
            self.assertDTypeEqual(dense_bounding_boxes["boxes"], tf.float16)
        else:
            self.assertDTypeEqual(outputs[IMAGES], tf.float16)

    @parameterized.named_parameters(*BUILD_IN_RUNTIME)
    def test_run_in_mixed_precision_and_build_in_runtime(
        self, layer_cls, args, build_layer_cls, build_args
    ):
        images = tf.cast(
            tf.random.uniform(shape=(4, 224, 224, 3)) * 255.0,
            dtype=tf.float64,
        )
        labels = tf.cast(
            tf.random.uniform(shape=(4, 1)) * 10.0, dtype=tf.float64
        )
        name = args["name"]
        value = args["value"]
        other_args = args["args"]
        if value == "single":
            build_layers = build_layer_cls(**build_args)
        elif value == "multiple":
            build_layers = [
                build_layer_cls(**build_args),
                build_layer_cls(**build_args),
            ]
        else:
            raise NotImplementedError()
        args = {name: build_layers, **other_args}
        layer = layer_cls(**args)

        outputs = layer({IMAGES: images, LABELS: labels})

        self.assertDTypeEqual(outputs[IMAGES], tf.float16)

    @classmethod
    def tearDownClass(cls) -> None:
        # Do not affect other tests
        keras.mixed_precision.set_global_policy("float32")
        tf.debugging.disable_check_numerics()


class MixedPrecisionBFloat16Test(tf.test.TestCase, parameterized.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        keras.mixed_precision.set_global_policy("mixed_bfloat16")

    def test_all_2d_aug_layers_included(self):
        base_cls = layers.VectorizedBaseRandomLayer
        cls_spaces = [augmentation, preprocessing]
        all_2d_aug_layers = []
        for cls_space in cls_spaces:
            all_2d_aug_layers.extend(
                inspect.getmembers(cls_space, predicate=inspect.isclass)
            )
        all_2d_aug_layer_names = set(
            item[0]
            for item in all_2d_aug_layers
            if issubclass(item[1], base_cls)
        )

        general_names = set(item[0] for item in GENERAL_TESTS)
        build_names = set(item[0] for item in BUILD_IN_RUNTIME)
        all_test_names = general_names.union(build_names)

        for name in all_2d_aug_layer_names:
            self.assertIn(name, all_test_names, msg=f"{name} not found")

    @parameterized.named_parameters(*GENERAL_TESTS)
    def test_run_in_mixed_precision(self, layer_cls, args, is_bbox_compatible):
        images = tf.cast(
            tf.random.uniform(shape=(2, 224, 224, 3)) * 255.0,
            dtype=tf.float64,
        )
        labels = tf.cast(
            tf.random.uniform(shape=(2, 1)) * 10.0, dtype=tf.float64
        )
        bounding_boxes = {
            "boxes": tf.ragged.constant(
                [
                    [[10, 10, 20, 20], [100, 100, 150, 150]],
                    [[200, 200, 400, 400]],
                ],
                dtype=tf.float32,
            ),
            "classes": tf.ragged.constant([[0, 1], [2]], dtype=tf.float32),
        }
        if is_bbox_compatible:
            try:
                layer = layer_cls(**args, bounding_box_format="xyxy")
            except TypeError:
                layer = layer_cls(**args)
            inputs = {
                IMAGES: images,
                LABELS: labels,
                BOUNDING_BOXES: bounding_boxes,
            }
        else:
            layer = layer_cls(**args)
            inputs = {IMAGES: images, LABELS: labels}

        outputs = layer(inputs)

        if is_bbox_compatible:
            self.assertDTypeEqual(outputs[IMAGES], tf.bfloat16)
            dense_bounding_boxes = bounding_box.to_dense(
                outputs[BOUNDING_BOXES]
            )
            self.assertDTypeEqual(dense_bounding_boxes["boxes"], tf.bfloat16)
        else:
            self.assertDTypeEqual(outputs[IMAGES], tf.bfloat16)

    @parameterized.named_parameters(*BUILD_IN_RUNTIME)
    def test_run_in_mixed_precision_and_build_in_runtime(
        self, layer_cls, args, build_layer_cls, build_args
    ):
        images = tf.cast(
            tf.random.uniform(shape=(4, 224, 224, 3)) * 255.0,
            dtype=tf.float64,
        )
        labels = tf.cast(
            tf.random.uniform(shape=(4, 1)) * 10.0, dtype=tf.float64
        )
        name = args["name"]
        value = args["value"]
        other_args = args["args"]
        if value == "single":
            build_layers = build_layer_cls(**build_args)
        elif value == "multiple":
            build_layers = [
                build_layer_cls(**build_args),
                build_layer_cls(**build_args),
            ]
        else:
            raise NotImplementedError()
        args = {name: build_layers, **other_args}
        layer = layer_cls(**args)

        outputs = layer({IMAGES: images, LABELS: labels})

        self.assertDTypeEqual(outputs[IMAGES], tf.bfloat16)

    @classmethod
    def tearDownClass(cls) -> None:
        # Do not affect other tests
        keras.mixed_precision.set_global_policy("float32")



"""
Most of these codes come from KerasCV.
"""
import tensorflow as tf

from keras_aug.datapoints import bounding_box


def to_ragged(bounding_boxes, sentinel=-1, dtype=tf.float32):
    """converts a Dense padded bounding box `tf.Tensor` to a `tf.RaggedTensor`.

    Bounding boxes are ragged tensors in most use cases. Converting them to a
    dense tensor makes it easier to work with Tensorflow ecosystem.
    This function can be used to filter out the masked out bounding boxes by
    checking for padded sentinel value of the class_id axis of the
    bounding_boxes.

    Usage:
    ```python
    bounding_boxes = {
        "boxes": tf.constant([[2, 3, 4, 5], [0, 1, 2, 3]]),
        "classes": tf.constant([[-1, 1]]),
    }
    bounding_boxes = bounding_box.to_ragged(bounding_boxes)
    print(bounding_boxes)
    # {
    #     "boxes": [[0, 1, 2, 3]],
    #     "classes": [[1]]
    # }
    ```

    Args:
        bounding_boxes: a Tensor of bounding boxes. May be batched, or
            unbatched.
        sentinel: The value indicating that a bounding box does not exist at the
            current index, and the corresponding box is padding, defaults to -1.
        dtype: the data type to use for the underlying Tensors.
    Returns:
        dictionary of `tf.RaggedTensor` or 'tf.Tensor' containing the filtered
        bounding boxes.
    """
    info = bounding_box.validate_format(bounding_boxes)

    if info["ragged"]:
        return bounding_boxes

    boxes = bounding_boxes.get("boxes")
    classes = bounding_boxes.get("classes")
    confidence = bounding_boxes.get("confidence", None)

    mask = classes != sentinel

    boxes = tf.ragged.boolean_mask(boxes, mask)
    classes = tf.ragged.boolean_mask(classes, mask)
    if confidence is not None:
        confidence = tf.ragged.boolean_mask(confidence, mask)

    if isinstance(boxes, tf.Tensor):
        boxes = tf.RaggedTensor.from_tensor(boxes)

    if isinstance(classes, tf.Tensor) and len(classes.shape) > 1:
        classes = tf.RaggedTensor.from_tensor(classes)

    if confidence is not None:
        if isinstance(confidence, tf.Tensor) and len(confidence.shape) > 1:
            confidence = tf.RaggedTensor.from_tensor(confidence)

    result = bounding_boxes.copy()
    result["boxes"] = tf.cast(boxes, dtype)
    result["classes"] = tf.cast(classes, dtype)

    if confidence is not None:
        result["confidence"] = tf.cast(confidence, dtype)

    return result



from tensorflow import keras

from keras_aug.datapoints import bounding_box
from keras_aug.layers.base.vectorized_base_random_layer import (
    VectorizedBaseRandomLayer,
)
from keras_aug.utils.augmentation import BOUNDING_BOXES


@keras.utils.register_keras_serializable(package="keras_aug")
class SanitizeBoundingBox(VectorizedBaseRandomLayer):
    """Remove degenerate/invalid bounding boxes.

    Args:
        min_size (int): The minimum size of the smaller side of bounding boxes.
        bounding_box_format (str): The format of bounding boxes of input
            dataset. Refer
            https://github.com/james77777778/keras-aug/blob/main/keras_aug/datapoints/bounding_box/converter.py
            for more details on supported bounding box formats.

    References:
        - `torchvision <https://github.com/pytorch/vision>`_
    """

    def __init__(self, min_size, bounding_box_format=None, **kwargs):
        super().__init__(**kwargs)
        self.min_size = min_size
        self.bounding_box_format = bounding_box_format

    def augment_ragged_image(self, image, transformation, **kwargs):
        return image

    def augment_images(self, images, transformations, **kwargs):
        return images

    def augment_labels(self, labels, transformations, **kwargs):
        return labels

    def augment_bounding_boxes(
        self, bounding_boxes, transformations, images=None, **kwargs
    ):
        if self.bounding_box_format is None:
            raise ValueError(
                "`SanitizeBoundingBox()` was called with bounding boxes,"
                "but no `bounding_box_format` was specified in the constructor."
                "Please specify a bounding box format in the constructor. i.e."
                "`SanitizeBoundingBox(..., bounding_box_format='xyxy')`"
            )
        bounding_boxes = bounding_box.to_dense(bounding_boxes)
        bounding_boxes = bounding_box.sanitize_bounding_boxes(
            bounding_boxes,
            min_size=self.min_size,
            bounding_box_format=self.bounding_box_format,
            images=images,
        )
        return bounding_boxes

    def augment_segmentation_masks(
        self, segmentation_masks, transformations, **kwargs
    ):
        return segmentation_masks

    def augment_keypoints(self, keypoints, transformations, **kwargs):
        return keypoints

    def _batch_augment(self, inputs):
        self._validate_inputs(inputs)
        return super()._batch_augment(inputs)

    def _validate_inputs(self, inputs):
        bounding_boxes = inputs.get(BOUNDING_BOXES, None)
        if bounding_boxes is None:
            raise ValueError(
                "SanitizeBoundingBox expects `bounding_boxes` to be present "
                "in its inputs. For example, "
                "`layer({'images': images, 'bounding_boxes': bounding_boxes})`."
            )

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "min_size": self.min_size,
                "bounding_box_format": self.bounding_box_format,
            }
        )
        return config



import tensorflow as tf
from tensorflow import keras

from keras_aug.datapoints import bounding_box
from keras_aug.layers.base.vectorized_base_random_layer import (
    VectorizedBaseRandomLayer,
)
from keras_aug.utils import augmentation as augmentation_utils


@keras.utils.register_keras_serializable(package="keras_aug")
class RandomZoomAndCrop(VectorizedBaseRandomLayer):
    """RandomZoomAndCrop implements resize with scale distortion.

    RandomZoomAndCrop takes a three-step approach to size-distortion based image
    augmentation. This technique is specifically tuned for object detection
    pipelines. The layer takes an input of images and bounding boxes, both of
    which may be ragged. It outputs a dense image tensor, ready to feed to a
    model for training. As such this layer will commonly be the final step in an
    augmentation pipeline.

    The augmentation process is as follows:
    The image is first scaled according to a randomly sampled scale factor. The
    width and height of the image are then resized according to the sampled
    scale. This is done to introduce noise into the local scale of features in
    the image. A subset of the image is then cropped randomly according to
    ``(crop_height, crop_width)``. This crop is then padded to be
    ``(height, width)``. Bounding boxes are translated and scaled according to
    the random scaling and random cropping.

    Args:
        height (int): The height of result image.
        width (int): The width of result image.
        scale_factor (float|Sequence[float]|keras_aug.FactorSampler): The range
            of the scale factor that is used to scale the input image. When
            represented as a single float, the factor will be picked between
            ``[1.0 - lower, 1.0 + upper]``. To reproduce the results of the
            MaskRCNN paper pass ``(0.8, 1.25)``.
        crop_height (int, optional): The height of the image to crop from the
            scaled image. Defaults to ``height`` when not provided.
        crop_width (int, optional): The width of the image to crop from the
            scaled image. Defaults to ``width`` when not provided.
        interpolation (str, optional): The interpolation method. Supported values:
            ``"nearest", "bilinear", "bicubic", "area", "lanczos3", "lanczos5", "gaussian", "mitchellcubic"``.
            Defaults to ``"bilinear"``.
        antialias (bool, optional): Whether to use antialias. Defaults to
            ``False``.
        position (str, optional): The padding method.
            Supported values: ``"center", "top_left", "top_right", "bottom_left", "bottom_right", "random"``.
            Defaults to ``"center"``.
        padding_value (int|float, optional): The padding value.
            Defaults to ``0``.
        bounding_box_format (str, optional): The format of bounding
            boxes of input dataset. Refer
            https://github.com/james77777778/keras-aug/blob/main/keras_aug/datapoints/bounding_box/converter.py
            for more details on supported bounding box formats.
        seed (int|float, optional): The random seed. Defaults to ``None``.

    References:
        - `KerasCV <https://github.com/keras-team/keras-cv>`_
    """  # noqa: E501

    def __init__(
        self,
        height,
        width,
        scale_factor,
        crop_height=None,
        crop_width=None,
        interpolation="bilinear",
        antialias=False,
        position="center",
        padding_value=0,
        bounding_box_format=None,
        seed=None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        if not isinstance(height, int) or not isinstance(width, int):
            raise ValueError(
                "RandomZoomAndCrop() expects ``height`` and ``width`` to be "
                f"integers. Received ``height={height}, width={width}``"
            )
        self.height = height
        self.width = width
        self.crop_height = crop_height or height
        self.crop_width = crop_width or width
        self.scale_factor = augmentation_utils.parse_factor(
            scale_factor,
            min_value=0.0,
            max_value=None,
            center_value=1.0,
            seed=seed,
        )
        self.interpolation = augmentation_utils.get_interpolation(interpolation)
        self.antialias = antialias
        self.position = augmentation_utils.get_padding_position(position)
        self.padding_value = padding_value
        self.bounding_box_format = bounding_box_format
        self.seed = seed

        crop_size = tf.expand_dims(
            tf.stack([self.crop_height, self.crop_width]), axis=0
        )
        self.crop_size = tf.cast(crop_size, dtype=tf.float32)
        self.force_output_dense_images = True

    def compute_ragged_image_signature(self, images):
        ragged_spec = tf.RaggedTensorSpec(
            shape=(self.height, self.width, images.shape[-1]),
            ragged_rank=1,
            dtype=self.compute_dtype,
        )
        return ragged_spec

    def get_random_transformation_batch(
        self, batch_size, images=None, **kwargs
    ):
        # cast to float32 to avoid numerical issue
        heights, widths = augmentation_utils.get_images_shape(
            images, dtype=tf.float32
        )

        # image_scales
        image_shapes = tf.concat((heights, widths), axis=-1)
        scaled_sizes = tf.round(
            image_shapes
            * self.scale_factor(shape=(batch_size, 1), dtype=tf.float32)
        )
        scales = tf.where(
            tf.less(
                scaled_sizes[..., 0:1] / image_shapes[..., 0:1],
                scaled_sizes[..., 1:] / image_shapes[..., 1:],
            ),
            scaled_sizes[..., 0:1] / image_shapes[..., 0:1],
            scaled_sizes[..., 1:] / image_shapes[..., 1:],
        )
        scaled_sizes = tf.round(image_shapes * scales)
        image_scales = scaled_sizes / image_shapes

        # offsets
        max_offsets = scaled_sizes - self.crop_size
        max_offsets = tf.where(
            tf.less(max_offsets, 0), tf.zeros_like(max_offsets), max_offsets
        )
        offsets = max_offsets * self._random_generator.random_uniform(
            shape=(batch_size, 2), minval=0, maxval=1, dtype=tf.float32
        )
        offsets = tf.cast(offsets, dtype=tf.int32)

        # paddings
        new_heights = tf.cast(scaled_sizes[..., 0:1], dtype=tf.int32)
        new_widths = tf.cast(scaled_sizes[..., 1:], dtype=tf.int32)
        tops = tf.where(
            new_heights < self.crop_height,
            tf.cast((self.crop_height - new_heights) / 2, tf.int32),
            0,
        )
        bottoms = tf.where(
            new_heights < self.crop_height,
            self.crop_height - new_heights - tops,
            0,
        )
        lefts = tf.where(
            new_widths < self.crop_width,
            tf.cast((self.crop_width - new_widths) / 2, tf.int32),
            0,
        )
        rights = tf.where(
            new_widths < self.crop_width,
            self.crop_width - new_widths - lefts,
            0,
        )
        (tops, bottoms, lefts, rights) = augmentation_utils.get_position_params(
            tops, bottoms, lefts, rights, self.position, self._random_generator
        )
        paddings = tf.concat([tops, bottoms, lefts, rights], axis=-1)

        return {
            "image_scales": image_scales,
            "scaled_sizes": scaled_sizes,
            "offsets": offsets,
            "paddings": paddings,
        }

    def augment_ragged_image(self, image, transformation, **kwargs):
        image = tf.expand_dims(image, axis=0)
        transformation = augmentation_utils.expand_dict_dims(
            transformation, axis=0
        )
        image = self.augment_images(
            images=image, transformations=transformation, **kwargs
        )
        return tf.squeeze(image, axis=0)

    def augment_images(self, images, transformations, **kwargs):
        # tf.image.resize always output tf.float32 unless interpolation==nearest
        inputs_for_resize_and_crop_single_image = {
            "images": tf.cast(images, dtype=tf.float32),
            "scaled_sizes": transformations["scaled_sizes"],
            "offsets": transformations["offsets"],
            "paddings": transformations["paddings"],
        }
        images = tf.map_fn(
            self.resize_and_crop_single_image,
            inputs_for_resize_and_crop_single_image,
            fn_output_signature=tf.float32,
        )
        images = tf.ensure_shape(
            images, shape=(None, self.height, self.width, None)
        )
        return tf.cast(images, self.compute_dtype)

    def augment_labels(self, labels, transformations, **kwargs):
        return labels

    def augment_bounding_boxes(
        self,
        bounding_boxes,
        transformations,
        images=None,
        raw_images=None,
        **kwargs,
    ):
        if self.bounding_box_format is None:
            raise ValueError(
                "`RandomZoomAndCrop()` was called with bounding boxes,"
                "but no `bounding_box_format` was specified in the constructor."
                "Please specify a bounding box format in the constructor. i.e."
                "`RandomZoomAndCrop(..., bounding_box_format='xyxy')`"
            )
        # cast to float32 to avoid numerical issue
        bounding_boxes = bounding_box.to_dense(bounding_boxes)
        bounding_boxes = bounding_box.convert_format(
            bounding_boxes,
            source=self.bounding_box_format,
            target="yxyx",
            images=raw_images,
            dtype=tf.float32,
        )

        image_scales = tf.cast(
            transformations["image_scales"], dtype=tf.float32
        )
        offsets = tf.cast(transformations["offsets"], dtype=tf.float32)
        paddings = tf.cast(transformations["paddings"], dtype=tf.float32)
        padding_offsets = tf.concat(
            [paddings[..., 0:1], paddings[..., 2:3]], axis=-1
        )

        # Adjusts box coordinates based on image_scale and offset.
        bounding_boxes = bounding_boxes.copy()
        yxyx = bounding_boxes["boxes"]
        yxyx *= tf.tile(image_scales, [1, 2])[..., tf.newaxis, :]
        yxyx -= tf.tile(offsets, [1, 2])[..., tf.newaxis, :]
        yxyx += tf.tile(padding_offsets, [1, 2])[..., tf.newaxis, :]

        bounding_boxes["boxes"] = yxyx
        bounding_boxes = bounding_box.clip_to_image(
            bounding_boxes,
            bounding_box_format="yxyx",
            images=images,
        )
        bounding_boxes = bounding_box.convert_format(
            bounding_boxes,
            source="yxyx",
            target=self.bounding_box_format,
            images=images,
            dtype=self.compute_dtype,
        )
        return bounding_boxes

    def compute_ragged_segmentation_mask_signature(self, segmentation_masks):
        return tf.RaggedTensorSpec(
            shape=(self.height, self.width, segmentation_masks.shape[-1]),
            ragged_rank=1,
            dtype=self.compute_dtype,
        )

    def augment_ragged_segmentation_mask(
        self, segmentation_mask, transformation, **kwargs
    ):
        segmentation_mask = tf.expand_dims(segmentation_mask, axis=0)
        transformation = augmentation_utils.expand_dict_dims(
            transformation, axis=0
        )
        segmentation_mask = self.augment_segmentation_masks(
            segmentation_masks=segmentation_mask,
            transformations=transformation,
            **kwargs,
        )
        return tf.squeeze(segmentation_mask, axis=0)

    def augment_segmentation_masks(
        self, segmentation_masks, transformations, **kwargs
    ):
        # unpackage augmentation arguments
        inputs_for_resize_and_crop_single_segmentation_mask = {
            "segmentation_masks": segmentation_masks,
            "scaled_sizes": transformations["scaled_sizes"],
            "offsets": transformations["offsets"],
            "paddings": transformations["paddings"],
        }
        segmentation_masks = tf.map_fn(
            self.resize_and_crop_single_segmentation_mask,
            inputs_for_resize_and_crop_single_segmentation_mask,
            fn_output_signature=segmentation_masks.dtype,
        )
        segmentation_masks = tf.ensure_shape(
            segmentation_masks, shape=(None, self.height, self.width, None)
        )
        return tf.cast(segmentation_masks, self.compute_dtype)

    def resize_and_crop_single_image(self, inputs):
        image = inputs.get("images", None)
        scaled_size = inputs.get("scaled_sizes", None)
        offset = inputs.get("offsets", None)
        padding = inputs.get("paddings", None)

        image = tf.image.resize(
            image,
            tf.cast(scaled_size, tf.int32),
            method=self.interpolation,
            antialias=self.antialias,
        )
        image = image[
            offset[0] : offset[0] + self.crop_height,
            offset[1] : offset[1] + self.crop_width,
            :,
        ]
        paddings = tf.stack(
            (
                tf.stack((padding[0], padding[1])),
                tf.stack((padding[2], padding[3])),
                tf.zeros(shape=(2,), dtype=tf.int32),
            )
        )
        image = tf.pad(
            image, paddings=paddings, constant_values=self.padding_value
        )
        return image

    def resize_and_crop_single_segmentation_mask(self, inputs):
        segmentation_mask = inputs.get("segmentation_masks", None)
        scaled_size = inputs.get("scaled_sizes", None)
        offset = inputs.get("offsets", None)
        padding = inputs.get("paddings", None)

        segmentation_mask = tf.image.resize(
            segmentation_mask, tf.cast(scaled_size, tf.int32), method="nearest"
        )
        segmentation_mask = segmentation_mask[
            offset[0] : offset[0] + self.crop_height,
            offset[1] : offset[1] + self.crop_width,
            :,
        ]
        paddings = tf.stack(
            (
                tf.stack((padding[0], padding[1])),
                tf.stack((padding[2], padding[3])),
                tf.zeros(shape=(2,), dtype=tf.int32),
            )
        )
        segmentation_mask = tf.pad(
            segmentation_mask,
            paddings=paddings,
            constant_values=self.padding_value,
        )
        return segmentation_mask

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "height": self.height,
                "width": self.width,
                "scale_factor": self.scale_factor,
                "crop_height": self.crop_height,
                "crop_width": self.crop_width,
                "interpolation": self.interpolation,
                "bounding_box_format": self.bounding_box_format,
                "seed": self.seed,
            }
        )
        return config



from keras_aug import datapoints
from keras_aug import layers
from keras_aug import utils
from keras_aug.core import ConstantFactorSampler
from keras_aug.core import FactorSampler
from keras_aug.core import NormalFactorSampler
from keras_aug.core import SignedConstantFactorSampler
from keras_aug.core import SignedNormalFactorSampler
from keras_aug.core import UniformFactorSampler

__version__ = "0.5.7"



import numpy as np
import tensorflow as tf
from keras_aug.datapoints import bounding_box

from keras_aug import layers


class RandomAffineTest(tf.test.TestCase):
    regular_args = {
        "rotation_factor": 10,
        "translation_height_factor": 0.1,
        "translation_width_factor": 0.1,
        "zoom_height_factor": 0.1,
        "zoom_width_factor": 0.1,
        "shear_height_factor": 0.1,
        "shear_width_factor": 0.1,
        "fill_mode": "constant",
        "fill_value": 0,
        "interpolation": "bilinear",
        "bounding_box_format": "xyxy",
    }
    no_aug_args = {
        "rotation_factor": 0.0,
        "translation_height_factor": 0.0,
        "translation_width_factor": 0.0,
        "zoom_height_factor": 0.0,
        "zoom_width_factor": 0.0,
        "shear_height_factor": 0.0,
        "shear_width_factor": 0.0,
        "fill_mode": "constant",
        "fill_value": 0,
        "interpolation": "bilinear",
        "bounding_box_format": "xyxy",
    }

    def test_no_adjustment(self):
        image_shape = (4, 8, 8, 3)
        image = tf.random.uniform(shape=image_shape) * 255.0

        layer = layers.RandomAffine(**self.no_aug_args)
        output = layer(image)
        self.assertAllClose(image, output, atol=1e-5, rtol=1e-5)

    def test_adjustment_for_non_rgb_value_range(self):
        image_shape = (4, 8, 8, 3)
        # Value range (0, 100)
        image = tf.random.uniform(shape=image_shape) * 100.0

        layer = layers.RandomAffine(**self.no_aug_args)
        output = layer(image)
        self.assertAllClose(image, output, rtol=1e-5, atol=1e-5)

        layer = layers.RandomAffine(**self.regular_args)
        output = layer(image)
        self.assertNotAllClose(image, output)

    def test_rotation_unbatched_image(self):
        input_image = np.reshape(np.arange(0, 25), (5, 5, 1)).astype(np.float32)
        # 180 rotation
        args = self.no_aug_args.copy()
        args.update({"rotation_factor": (180, 180)})
        layer = layers.RandomAffine(**args)
        expected_output = np.asarray(
            [
                [24, 23, 22, 21, 20],
                [19, 18, 17, 16, 15],
                [14, 13, 12, 11, 10],
                [9, 8, 7, 6, 5],
                [4, 3, 2, 1, 0],
            ]
        ).astype(np.float32)

        output_image = layer(input_image)
        expected_output = np.reshape(expected_output, (5, 5, 1))

        self.assertAllClose(expected_output, output_image)

    def test_rotation_bounding_boxes(self):
        input_image = np.random.random((512, 512, 3)).astype(np.float32)
        bounding_boxes = {
            "boxes": tf.convert_to_tensor(
                [[200, 200, 400, 400], [100, 100, 300, 300]], dtype=tf.float32
            ),
            "classes": tf.convert_to_tensor([1, 2], dtype=tf.float32),
        }
        input = {"images": input_image, "bounding_boxes": bounding_boxes}
        expected_bounding_boxes = {
            "boxes": tf.convert_to_tensor(
                [[112.0, 112.0, 312.0, 312.0], [212.0, 212.0, 412.0, 412.0]],
                dtype=tf.float32,
            ),
            "classes": tf.convert_to_tensor([1, 2], dtype=tf.float32),
        }
        # 180 rotation
        args = self.no_aug_args.copy()
        args.update({"rotation_factor": (180, 180)})
        layer = layers.RandomAffine(**args)

        output = layer(input)
        output["bounding_boxes"] = bounding_box.to_dense(
            output["bounding_boxes"]
        )

        self.assertAllClose(expected_bounding_boxes, output["bounding_boxes"])

    def test_rotation_ragged_bounding_boxes(self):
        input_image = np.random.random((2, 512, 512, 3)).astype(np.float32)
        bounding_boxes = {
            "boxes": tf.ragged.constant(
                [
                    [[200, 200, 400, 400], [100, 100, 300, 300]],
                    [[200, 200, 400, 400]],
                ],
                dtype=tf.float32,
            ),
            "classes": tf.ragged.constant(
                [[0, 0], [0]],
                dtype=tf.float32,
            ),
        }
        input = {"images": input_image, "bounding_boxes": bounding_boxes}
        expected_output = {
            "boxes": tf.ragged.constant(
                [
                    [
                        [112.0, 112.0, 312.0, 312.0],
                        [212.0, 212.0, 412.0, 412.0],
                    ],
                    [[112.0, 112.0, 312.0, 312.0]],
                ],
                dtype=tf.float32,
            ),
            "classes": tf.ragged.constant(
                [[0, 0], [0]],
                dtype=tf.float32,
            ),
        }
        # 180 rotation
        args = self.no_aug_args.copy()
        args.update({"rotation_factor": (180, 180)})
        layer = layers.RandomAffine(**args)

        output = layer(input)

        expected_output = bounding_box.to_dense(expected_output)
        output["bounding_boxes"] = bounding_box.to_dense(
            output["bounding_boxes"]
        )

        self.assertAllClose(
            expected_output["boxes"], output["bounding_boxes"]["boxes"]
        )
        self.assertAllClose(
            expected_output["classes"],
            output["bounding_boxes"]["classes"],
        )

    def test_rotation_sparse_segmentation_mask(self):
        num_classes = 8

        input_images = np.random.random((2, 20, 20, 3)).astype(np.float32)
        # Masks are all 0s or 8s, to verify that when we rotate we don't do bad
        # mask interpolation to either a 0 or a 7
        masks = np.random.randint(2, size=(2, 20, 20, 1)) * (num_classes - 1)
        inputs = {"images": input_images, "segmentation_masks": masks}

        # 90 rotation
        args = self.no_aug_args.copy()
        args.update({"rotation_factor": (90, 90)})
        layer = layers.RandomAffine(**args)
        outputs = layer(inputs)
        expected_masks = np.rot90(masks, axes=(1, 2))
        self.assertAllClose(expected_masks, outputs["segmentation_masks"])

        # 45-degree rotation. Only verifies that no interpolation takes place.
        # 90 rotation
        args = self.no_aug_args.copy()
        args.update({"rotation_factor": (0.125, 0.125)})
        outputs = layer(inputs)
        self.assertAllInSet(outputs["segmentation_masks"], [0, 7])

    def test_translation_up_constant(self):
        for dtype in (np.int64, np.float32):
            input_image = np.reshape(np.arange(0, 25), (1, 5, 5, 1)).astype(
                dtype
            )
            # Shifting by -.2 * 5 = 1 pixel.
            args = self.no_aug_args.copy()
            args.update({"translation_height_factor": (-0.2, -0.2)})
            layer = layers.RandomAffine(**args)
            expected_output = (
                np.asarray(
                    [
                        [5, 6, 7, 8, 9],
                        [10, 11, 12, 13, 14],
                        [15, 16, 17, 18, 19],
                        [20, 21, 22, 23, 24],
                        [0, 0, 0, 0, 0],
                    ]
                )
                .astype(dtype)
                .reshape((1, 5, 5, 1))
            )

            output_image = layer(input_image)

            self.assertAllEqual(expected_output, output_image)

    def test_translation_down_constant(self):
        for dtype in (np.int64, np.float32):
            input_image = np.reshape(np.arange(0, 25), (1, 5, 5, 1)).astype(
                dtype
            )
            # Shifting by -.2 * 5 = 1 pixel.
            args = self.no_aug_args.copy()
            args.update({"translation_height_factor": (0.2, 0.2)})
            layer = layers.RandomAffine(**args)
            expected_output = (
                np.asarray(
                    [
                        [0, 0, 0, 0, 0],
                        [0, 1, 2, 3, 4],
                        [5, 6, 7, 8, 9],
                        [10, 11, 12, 13, 14],
                        [15, 16, 17, 18, 19],
                    ]
                )
                .astype(dtype)
                .reshape((1, 5, 5, 1))
            )

            output_image = layer(input_image)

            self.assertAllEqual(expected_output, output_image)

    def test_translation_left_constant(self):
        for dtype in (np.int64, np.float32):
            input_image = np.reshape(np.arange(0, 25), (1, 5, 5, 1)).astype(
                dtype
            )
            # Shifting by -.2 * 5 = 1 pixel.
            args = self.no_aug_args.copy()
            args.update({"translation_width_factor": (-0.2, -0.2)})
            layer = layers.RandomAffine(**args)
            expected_output = (
                np.asarray(
                    [
                        [1, 2, 3, 4, 0],
                        [6, 7, 8, 9, 0],
                        [11, 12, 13, 14, 0],
                        [16, 17, 18, 19, 0],
                        [21, 22, 23, 24, 0],
                    ]
                )
                .astype(dtype)
                .reshape((1, 5, 5, 1))
            )

            output_image = layer(input_image)

            self.assertAllEqual(expected_output, output_image)

    def test_translation_right_constant(self):
        for dtype in (np.int64, np.float32):
            input_image = np.reshape(np.arange(0, 25), (1, 5, 5, 1)).astype(
                dtype
            )
            # Shifting by -.2 * 5 = 1 pixel.
            args = self.no_aug_args.copy()
            args.update({"translation_width_factor": (0.2, 0.2)})
            layer = layers.RandomAffine(**args)
            expected_output = (
                np.asarray(
                    [
                        [0, 0, 1, 2, 3],
                        [0, 5, 6, 7, 8],
                        [0, 10, 11, 12, 13],
                        [0, 15, 16, 17, 18],
                        [0, 20, 21, 22, 23],
                    ]
                )
                .astype(dtype)
                .reshape((1, 5, 5, 1))
            )

            output_image = layer(input_image)

            self.assertAllEqual(expected_output, output_image)

    def test_zoom_in_numeric(self):
        for dtype in (np.int64, np.float32):
            input_image = np.reshape(np.arange(0, 25), (5, 5, 1)).astype(dtype)
            args = self.no_aug_args.copy()
            args.update(
                {
                    "zoom_height_factor": (0.5, 0.5),
                    "zoom_width_factor": (0.5, 0.5),
                    "interpolation": "nearest",
                }
            )
            layer = layers.RandomAffine(**args)
            expected_output = (
                np.asarray(
                    [
                        [6, 7, 7, 8, 8],
                        [11, 12, 12, 13, 13],
                        [11, 12, 12, 13, 13],
                        [16, 17, 17, 18, 18],
                        [16, 17, 17, 18, 18],
                    ]
                )
                .astype(dtype)
                .reshape((1, 5, 5, 1))
            )

            output_image = layer(np.expand_dims(input_image, axis=0))

            self.assertAllEqual(expected_output, output_image)

    def test_zoom_out_numeric(self):
        for dtype in (np.int64, np.float32):
            input_image = np.reshape(np.arange(0, 25), (5, 5, 1)).astype(dtype)
            args = self.no_aug_args.copy()
            args.update(
                {
                    "zoom_height_factor": (1.5, 1.5),
                    "zoom_width_factor": (1.8, 1.8),
                    "interpolation": "nearest",
                }
            )
            layer = layers.RandomAffine(**args)
            expected_output = (
                np.asarray(
                    [
                        [0, 0, 0, 0, 0],
                        [0, 5, 7, 9, 0],
                        [0, 10, 12, 14, 0],
                        [0, 20, 22, 24, 0],
                        [0, 0, 0, 0, 0],
                    ]
                )
                .astype(dtype)
                .reshape((1, 5, 5, 1))
            )

            output_image = layer(np.expand_dims(input_image, axis=0))

            self.assertAllEqual(expected_output, output_image)

    def test_shear_area(self):
        xs = tf.ones((1, 512, 512, 3))
        ys = {
            "boxes": tf.constant(
                [[[0.3, 0.4, 0.5, 0.6], [0.9, 0.8, 1.0, 1.0]]]
            ),
            "classes": tf.constant([[2, 3]]),
        }

        inputs = {"images": xs, "bounding_boxes": ys}
        args = self.no_aug_args.copy()
        args.update(
            {
                "shear_height_factor": (0.3, 0.7),
                "shear_width_factor": (0.4, 0.7),
                "interpolation": "nearest",
                "seed": 0,
                "bounding_box_format": "rel_xyxy",
            }
        )
        layer = layers.RandomAffine(**args)

        outputs = layer(inputs)
        xs, ys_bounding_boxes = (
            outputs["images"],
            outputs["bounding_boxes"]["boxes"],
        )
        new_area = tf.math.multiply(
            tf.abs(
                tf.subtract(
                    ys_bounding_boxes[..., 2], ys_bounding_boxes[..., 0]
                )
            ),
            tf.abs(
                tf.subtract(
                    ys_bounding_boxes[..., 3], ys_bounding_boxes[..., 1]
                )
            ),
        )
        old_area = tf.math.multiply(
            tf.abs(tf.subtract(ys["boxes"][..., 2], ys["boxes"][..., 0])),
            tf.abs(tf.subtract(ys["boxes"][..., 3], ys["boxes"][..., 1])),
        )

        self.assertTrue(tf.math.reduce_all(new_area > old_area))

    def test_dense_segmentation_masks(self):
        images = tf.random.uniform((2, 10, 10, 3))
        segmentation_masks = tf.random.uniform(
            (2, 10, 10, 1), minval=0, maxval=10, dtype=tf.int32
        )
        args = self.no_aug_args.copy()
        args.update(
            {
                "shear_height_factor": (0.3, 0.7),
                "shear_width_factor": (0.4, 0.7),
                "interpolation": "nearest",
                "seed": 0,
            }
        )
        layer = layers.RandomAffine(**args)

        result = layer(
            {"images": images, "segmentation_masks": segmentation_masks}
        )

        self.assertTrue(isinstance(result["segmentation_masks"], tf.Tensor))

    def test_ragged_segmentation_masks(self):
        images = tf.ragged.stack(
            [
                tf.random.uniform((8, 8, 3), dtype=tf.float32),
                tf.random.uniform((16, 8, 3), dtype=tf.float32),
            ]
        )
        segmentation_masks = tf.ragged.stack(
            [
                tf.random.uniform((8, 8, 1), maxval=10, dtype=tf.int32),
                tf.random.uniform((16, 8, 1), maxval=10, dtype=tf.int32),
            ]
        )
        segmentation_masks = tf.cast(segmentation_masks, dtype=tf.float32)
        args = self.no_aug_args.copy()
        args.update(
            {
                "shear_height_factor": (0.3, 0.7),
                "shear_width_factor": (0.4, 0.7),
                "interpolation": "nearest",
                "seed": 0,
            }
        )
        layer = layers.RandomAffine(**args)

        result = layer(
            {"images": images, "segmentation_masks": segmentation_masks}
        )

        self.assertTrue(
            isinstance(result["segmentation_masks"], tf.RaggedTensor)
        )



import tensorflow as tf
from absl.testing import parameterized

from keras_aug import layers


class RandomColorJitterTest(tf.test.TestCase, parameterized.TestCase):
    regular_args = {
        "value_range": (0, 255),
        "brightness_factor": (0.6, 1.4),  # 40%
        "contrast_factor": (0.6, 1.4),  # 40%
        "hue_factor": (-0.015, 0.015),  # 1.5%
        "saturation_factor": (0.6, 1.4),  # 40%
    }
    no_aug_args = {
        "value_range": (0, 255),
        "brightness_factor": (1.0, 1.0),
        "contrast_factor": (1.0, 1.0),
        "hue_factor": (0.0, 0.0),
        "saturation_factor": (1.0, 1.0),
    }

    def test_no_adjustment(self):
        image_shape = (4, 8, 8, 3)
        image = tf.random.uniform(shape=image_shape) * 255.0

        layer = layers.RandomColorJitter(**self.no_aug_args)
        output = layer(image)
        self.assertAllClose(image, output, atol=1e-5, rtol=1e-5)

    def test_adjustment_for_non_rgb_value_range(self):
        image_shape = (4, 8, 8, 3)
        # Value range (0, 100)
        image = tf.random.uniform(shape=image_shape) * 100.0

        layer = layers.RandomColorJitter(**self.no_aug_args)
        output = layer(image)
        self.assertAllClose(image, output, rtol=1e-5, atol=1e-5)

        layer = layers.RandomColorJitter(**self.regular_args)
        output = layer(image)
        self.assertNotAllClose(image, output)

    def test_max_brightness(self):
        image_shape = (4, 8, 8, 3)
        image = tf.random.uniform(shape=image_shape, seed=2023) * 255.0
        args = self.no_aug_args.copy()
        args.update({"brightness_factor": (10000.0, 10000.0)})
        layer = layers.RandomColorJitter(**args, seed=2023)

        output = layer(image)

        self.assertAllClose(output, tf.fill((4, 8, 8, 3), 255))

    def test_max_brightness_rescaled_value_range(self):
        image_shape = (4, 8, 8, 3)
        image = tf.random.uniform(shape=image_shape, seed=2023)
        args = self.no_aug_args.copy()
        args.update(
            {"value_range": (0, 1), "brightness_factor": (10000.0, 10000.0)}
        )
        layer = layers.RandomColorJitter(**args, seed=2023)

        output = layer(image)

        self.assertAllClose(output, tf.fill((4, 8, 8, 3), 1))

    def test_zero_brightness(self):
        image_shape = (4, 8, 8, 3)
        image = tf.random.uniform(shape=image_shape) * 255.0

        args = self.no_aug_args.copy()
        args.update({"brightness_factor": (0.0, 0.0)})
        layer = layers.RandomColorJitter(**args)

        output = layer(image)

        self.assertAllClose(output, tf.fill((4, 8, 8, 3), 0))

    def test_adjust_full_opposite_hue(self):
        image_shape = (4, 8, 8, 3)
        image = tf.random.uniform(shape=image_shape) * 255.0
        args = self.no_aug_args.copy()
        args.update({"hue_factor": (0.0, 0.0)})

        layer = layers.RandomColorJitter(**args)
        output = layer(image)

        channel_max = tf.math.reduce_max(output, axis=-1)
        channel_min = tf.math.reduce_min(output, axis=-1)
        # Make sure the max and min channel are the same between input and
        # output. In the meantime, and channel will swap between each other.
        self.assertAllClose(
            channel_max,
            tf.math.reduce_max(image, axis=-1),
            atol=1e-5,
            rtol=1e-5,
        )
        self.assertAllClose(
            channel_min,
            tf.math.reduce_min(image, axis=-1),
            atol=1e-5,
            rtol=1e-5,
        )

    def test_adjust_to_grayscale(self):
        image_shape = (4, 8, 8, 3)
        image = tf.random.uniform(shape=image_shape) * 255.0

        args = self.no_aug_args.copy()
        args.update({"saturation_factor": (0.0, 0.0)})
        layer = layers.RandomColorJitter(**args)
        output = layer(image)

        channel_mean = tf.math.reduce_mean(output, axis=-1)
        channel_values = tf.unstack(output, axis=-1)
        # Make sure all the pixel has the same value among the channel dim,
        # which is a fully gray RGB.
        for channel_value in channel_values:
            self.assertAllClose(
                channel_mean, channel_value, atol=1e-5, rtol=1e-5
            )

    def test_adjust_to_full_saturation(self):
        image_shape = (4, 8, 8, 3)
        image = tf.random.uniform(shape=image_shape) * 255.0

        args = self.no_aug_args.copy()
        args.update({"saturation_factor": (100.0, 100.0)})
        layer = layers.RandomColorJitter(**args)
        output = layer(image)

        channel_mean = tf.math.reduce_min(output, axis=-1)
        # Make sure at least one of the channel is 0.0 (fully saturated image)
        self.assertAllClose(channel_mean, tf.zeros((4, 8, 8)))

    @parameterized.named_parameters(("025", 0.25), ("05", 0.5))
    def test_adjusts_all_values_for_hue_factor(self, factor):
        image_shape = (4, 8, 8, 3)
        image = tf.random.uniform(shape=image_shape) * 255.0

        args = self.no_aug_args.copy()
        args.update({"hue_factor": (0.0 - factor, 0.0 + factor)})
        layer = layers.RandomColorJitter(**args)
        output = layer(image)
        self.assertNotAllClose(image, output)

    @parameterized.named_parameters(("025", 0.25), ("05", 0.5))
    def test_adjusts_all_values_for_saturation_factor(self, factor):
        image_shape = (4, 8, 8, 3)
        image = tf.random.uniform(shape=image_shape) * 255.0

        args = self.no_aug_args.copy()
        args.update({"saturation_factor": (1.0 - factor, 1.0 + factor)})
        layer = layers.RandomColorJitter(**args)
        output = layer(image)
        self.assertNotAllClose(image, output)



import tensorflow as tf

from keras_aug import layers


class MosaicTest(tf.test.TestCase):
    num_classes = 10

    def test_return_shapes(self):
        xs = tf.ones((2, 32, 32, 3))
        # randomly sample labels
        ys_labels = tf.random.categorical(tf.math.log([[0.5, 0.5]]), 2)
        ys_labels = tf.squeeze(ys_labels)
        ys_labels = tf.one_hot(ys_labels, self.num_classes)

        # randomly sample bounding boxes
        ys_bounding_boxes = {
            "boxes": tf.random.uniform((2, 3, 4), 0, 1),
            "classes": tf.random.uniform((2, 3), 0, 1),
        }
        layer = layers.Mosaic(height=32, width=32, bounding_box_format="xywh")
        # mosaic on labels
        outputs = layer(
            {
                "images": xs,
                "labels": ys_labels,
                "bounding_boxes": ys_bounding_boxes,
            }
        )
        xs, ys_labels, ys_bounding_boxes = (
            outputs["images"],
            outputs["labels"],
            outputs["bounding_boxes"],
        )

        self.assertEqual(xs.shape, [2, 32, 32, 3])
        self.assertEqual(ys_labels.shape, [2, 10])
        self.assertEqual(ys_bounding_boxes["boxes"].shape, [2, 3 * 4, 4])
        self.assertEqual(ys_bounding_boxes["classes"].shape, [2, 3 * 4])

    def test_image_input_only(self):
        xs = tf.cast(
            tf.stack([2 * tf.ones((10, 10, 1)), tf.ones((10, 10, 1))], axis=0),
            tf.float32,
        )
        layer = layers.Mosaic(height=32, width=32)
        with self.assertRaisesRegexp(
            ValueError, "expects inputs in a dictionary"
        ):
            _ = layer(xs)

    def test_single_image_input(self):
        xs = tf.ones((32, 32, 3))
        ys = tf.one_hot(tf.constant([1]), 2)
        inputs = {"images": xs, "labels": ys}
        layer = layers.Mosaic(height=32, width=32)
        with self.assertRaisesRegexp(
            ValueError,
            "Mosaic received a single image to `call`",
        ):
            _ = layer(inputs)

    def test_image_input(self):
        xs = tf.ones((2, 32, 32, 3))
        layer = layers.Mosaic(height=32, width=32)
        with self.assertRaisesRegexp(
            ValueError,
            "Mosaic expects inputs in a dictionary with format",
        ):
            _ = layer(xs)

    def test_ragged_input_with_graph_mode(self):
        images = tf.ragged.stack(
            [
                tf.random.uniform((8, 8, 3), dtype=tf.float32),
                tf.random.uniform((16, 8, 3), dtype=tf.float32),
                tf.random.uniform((8, 8, 3), dtype=tf.float32),
                tf.random.uniform((16, 8, 3), dtype=tf.float32),
            ]
        )
        # randomly sample labels
        labels = tf.random.categorical(tf.math.log([[0.5, 0.5, 0.5, 0.5]]), 2)
        labels = tf.squeeze(labels)
        labels = tf.one_hot(labels, 10)
        segmentation_masks = tf.ragged.stack(
            [
                tf.random.uniform((8, 8, 1), maxval=10, dtype=tf.int32),
                tf.random.uniform((16, 8, 1), maxval=10, dtype=tf.int32),
                tf.random.uniform((8, 8, 1), maxval=10, dtype=tf.int32),
                tf.random.uniform((16, 8, 1), maxval=10, dtype=tf.int32),
            ]
        )
        layer = layers.Mosaic(height=8, width=8)

        @tf.function
        def fn(inputs):
            outputs = layer(inputs)
            image_shape = outputs["images"].shape
            segmentation_mask_shape = outputs["segmentation_masks"].shape
            self.assertEqual(image_shape, (4, 8, 8, 3))
            self.assertEqual(segmentation_mask_shape, (4, 8, 8, 1))

        fn(
            {
                "images": images,
                "labels": labels,
                "segmentation_masks": segmentation_masks,
            }
        )



import tensorflow as tf

from keras_aug import layers


class RandomChannelDropoutTest(tf.test.TestCase):
    regular_args = {"factor": (0, 2)}

    def test_layer_created_with_invalid_factor(self):
        images = tf.ones(shape=(2, 32, 32, 3))
        args = self.regular_args.copy()
        args.update({"factor": (5, 5)})
        layer = layers.RandomChannelDropout(**args)

        with self.assertRaises(ValueError):
            _ = layer(images)

    def test_partially_zeros_out(self):
        images = tf.ones(shape=(2, 32, 32, 3))
        layer = layers.RandomChannelDropout(**self.regular_args)

        results = layer(images)

        self.assertEqual(tf.reduce_sum(results), 2 * 32 * 32 * 2)



import tensorflow as tf
from tensorflow import keras

from keras_aug.datapoints import bounding_box
from keras_aug.layers.base.vectorized_base_random_layer import (
    VectorizedBaseRandomLayer,
)
from keras_aug.utils import augmentation as augmentation_utils


@keras.utils.register_keras_serializable(package="keras_aug")
class RandomCrop(VectorizedBaseRandomLayer):
    """Randomly crops the input images.

    This layer will randomly choose a location to crop images down to
    ``(height, width)``. If an input image is smaller than the
    ``(height, width)``, the input will be resized and cropped to return the
    largest possible window in the image that matches the target aspect ratio.

    Args:
        height (int): The height of result image.
        width (int): The width of result image.
        interpolation (str, optional): The interpolation mode. Supported values:
            ``"nearest", "bilinear"``. Defaults to `"bilinear"`.
        bounding_box_format (str, optional): The format of bounding
            boxes of input dataset. Refer
            https://github.com/james77777778/keras-aug/blob/main/keras_aug/datapoints/bounding_box/converter.py
            for more details on supported bounding box formats.
        bounding_box_min_area_ratio (float, optional): The threshold to
            apply sanitize_bounding_boxes. Defaults to ``None``.
        bounding_box_max_aspect_ratio (float, optional): The threshold to
            apply sanitize_bounding_boxes. Defaults to ``None``.
        seed (int|float, optional): The random seed. Defaults to ``None``.

    References:
        - `KerasCV <https://github.com/keras-team/keras-cv>`_
    """  # noqa: E501

    def __init__(
        self,
        height,
        width,
        interpolation="bilinear",
        bounding_box_format=None,
        bounding_box_min_area_ratio=None,
        bounding_box_max_aspect_ratio=None,
        seed=None,
        **kwargs,
    ):
        super().__init__(seed=seed, **kwargs)
        self.height = height
        self.width = width
        self.interpolation = interpolation
        self.bounding_box_format = bounding_box_format
        self.bounding_box_min_area_ratio = bounding_box_min_area_ratio
        self.bounding_box_max_aspect_ratio = bounding_box_max_aspect_ratio
        self.seed = seed

        self.resize_bilinear = keras.layers.Resizing(self.height, self.width)
        self.resize_nearest = keras.layers.Resizing(
            self.height, self.width, interpolation="nearest"
        )
        # set force_output_dense_images=True because the output images must
        # have same shape (B, height, width, C)
        self.force_output_dense_images = True

    def get_random_transformation_batch(
        self, batch_size, images=None, **kwargs
    ):
        # cast to float32 to avoid numerical issue
        tops = self._random_generator.random_uniform(
            shape=(batch_size, 1), minval=0, maxval=1, dtype=tf.float32
        )
        lefts = self._random_generator.random_uniform(
            shape=(batch_size, 1), minval=0, maxval=1, dtype=tf.float32
        )
        return {"crop_tops": tops, "crop_lefts": lefts}

    def compute_ragged_image_signature(self, images):
        ragged_spec = tf.RaggedTensorSpec(
            shape=(self.height, self.width, images.shape[-1]),
            ragged_rank=1,
            dtype=self.compute_dtype,
        )
        return ragged_spec

    def augment_ragged_image(self, image, transformation, **kwargs):
        image = tf.expand_dims(image, axis=0)
        transformation = augmentation_utils.expand_dict_dims(
            transformation, axis=0
        )
        image = self.augment_images(
            images=image, transformations=transformation, **kwargs
        )
        return tf.squeeze(image, axis=0)

    def augment_images(self, images, transformations, **kwargs):
        heights, widths = augmentation_utils.get_images_shape(images)
        h_diffs = heights - self.height
        w_diffs = widths - self.width
        # broadcast
        h_diffs = h_diffs[:, tf.newaxis, tf.newaxis, :]
        w_diffs = w_diffs[:, tf.newaxis, tf.newaxis, :]
        images = tf.where(
            tf.math.logical_and(h_diffs >= 0, w_diffs >= 0),
            self.crop_images(images, transformations),
            self.resize_images(images, interpolation=self.interpolation),
        )
        images = tf.ensure_shape(
            images, shape=(None, self.height, self.width, None)
        )
        return tf.cast(images, dtype=self.compute_dtype)

    def augment_labels(self, labels, transformations, **kwargs):
        return labels

    def augment_bounding_boxes(
        self,
        bounding_boxes,
        transformations,
        images=None,
        raw_images=None,
        **kwargs,
    ):
        if self.bounding_box_format is None:
            raise ValueError(
                "`RandomCrop()` was called with bounding boxes,"
                "but no `bounding_box_format` was specified in the constructor."
                "Please specify a bounding box format in the constructor. i.e."
                "`RandomCrop(bounding_box_format='xyxy')`"
            )
        bounding_boxes = bounding_box.to_dense(bounding_boxes)
        bounding_boxes = bounding_box.convert_format(
            bounding_boxes,
            source=self.bounding_box_format,
            target="xyxy",
            images=raw_images,
            dtype=self.compute_dtype,
        )
        original_bounding_boxes = bounding_boxes.copy()

        heights, widths = augmentation_utils.get_images_shape(raw_images)
        h_diffs = heights - self.height
        w_diffs = widths - self.width
        # broadcast
        h_diffs = h_diffs[:, tf.newaxis, :]
        w_diffs = w_diffs[:, tf.newaxis, :]
        boxes = tf.where(
            tf.math.logical_and(h_diffs >= 0, w_diffs >= 0),
            self.crop_bounding_boxes(
                raw_images, bounding_boxes["boxes"], transformations
            ),
            self.resize_bounding_boxes(
                raw_images,
                bounding_boxes["boxes"],
            ),
        )
        bounding_boxes["boxes"] = boxes
        bounding_boxes = bounding_box.clip_to_image(
            bounding_boxes,
            bounding_box_format="xyxy",
            images=images,
        )
        bounding_boxes = bounding_box.sanitize_bounding_boxes(
            bounding_boxes,
            min_area_ratio=self.bounding_box_min_area_ratio,
            max_aspect_ratio=self.bounding_box_max_aspect_ratio,
            bounding_box_format="xyxy",
            reference_bounding_boxes=original_bounding_boxes,
            images=images,
            reference_images=raw_images,
        )
        bounding_boxes = bounding_box.convert_format(
            bounding_boxes,
            source="xyxy",
            target=self.bounding_box_format,
            dtype=self.compute_dtype,
            images=images,
        )
        return bounding_boxes

    def compute_ragged_segmentation_mask_signature(self, segmentation_masks):
        return tf.RaggedTensorSpec(
            shape=(self.height, self.width, segmentation_masks.shape[-1]),
            ragged_rank=1,
            dtype=self.compute_dtype,
        )

    def augment_ragged_segmentation_mask(
        self, segmentation_mask, transformation, **kwargs
    ):
        segmentation_mask = tf.expand_dims(segmentation_mask, axis=0)
        transformation = augmentation_utils.expand_dict_dims(
            transformation, axis=0
        )
        segmentation_mask = self.augment_segmentation_masks(
            segmentation_masks=segmentation_mask,
            transformations=transformation,
            **kwargs,
        )
        return tf.squeeze(segmentation_mask, axis=0)

    def augment_segmentation_masks(
        self, segmentation_masks, transformations, **kwargs
    ):
        heights, widths = augmentation_utils.get_images_shape(
            segmentation_masks
        )
        h_diffs = heights - self.height
        w_diffs = widths - self.width
        # broadcast
        h_diffs = h_diffs[:, tf.newaxis, tf.newaxis, :]
        w_diffs = w_diffs[:, tf.newaxis, tf.newaxis, :]
        segmentation_masks = tf.where(
            tf.math.logical_and(h_diffs >= 0, w_diffs >= 0),
            self.crop_images(segmentation_masks, transformations),
            self.resize_images(segmentation_masks, interpolation="nearest"),
        )
        segmentation_masks = tf.ensure_shape(
            segmentation_masks, shape=(None, self.height, self.width, None)
        )
        return tf.cast(segmentation_masks, dtype=self.compute_dtype)

    def crop_images(self, images, transformations):
        batch_size = tf.shape(images)[0]
        heights, widths = augmentation_utils.get_images_shape(
            images, dtype=tf.float32
        )
        tops = transformations["crop_tops"]
        lefts = transformations["crop_lefts"]
        x1s = lefts * (widths - self.width)
        y1s = tops * (heights - self.height)
        x2s = x1s + self.width
        y2s = y1s + self.height
        # normalize
        x1s /= widths
        y1s /= heights
        x2s /= widths
        y2s /= heights
        boxes = tf.concat([y1s, x1s, y2s, x2s], axis=-1)

        # tf.image.crop_and_resize not support bfloat16
        if images.dtype == tf.bfloat16:
            images = tf.cast(images, dtype=tf.float32)
        images = tf.image.crop_and_resize(
            images,
            boxes,
            tf.range(batch_size),
            [self.height, self.width],
            method="nearest",
        )
        return tf.cast(images, dtype=self.compute_dtype)

    def resize_images(self, images, interpolation="bilinear"):
        if interpolation == "bilinear":
            images = self.resize_bilinear(images)
        elif interpolation == "nearest":
            images = self.resize_nearest(images)
        else:
            raise ValueError(f"Unsupported interpolation: {interpolation}")
        return tf.cast(images, dtype=self.compute_dtype)

    def crop_bounding_boxes(self, images, boxes, transformation):
        tops = transformation["crop_tops"]
        lefts = transformation["crop_lefts"]
        heights, widths = augmentation_utils.get_images_shape(
            images, dtype=tf.float32
        )

        # compute offsets for xyxy bounding_boxes
        top_offsets = tf.cast(
            tf.math.round(tops * (heights - self.height)),
            dtype=self.compute_dtype,
        )
        left_offsets = tf.cast(
            tf.math.round(lefts * (widths - self.width)),
            dtype=self.compute_dtype,
        )
        x1s, y1s, x2s, y2s = tf.split(boxes, 4, axis=-1)
        x1s -= tf.expand_dims(left_offsets, axis=1)
        y1s -= tf.expand_dims(top_offsets, axis=1)
        x2s -= tf.expand_dims(left_offsets, axis=1)
        y2s -= tf.expand_dims(top_offsets, axis=1)
        outputs = tf.concat([x1s, y1s, x2s, y2s], axis=-1)
        return outputs

    def resize_bounding_boxes(self, images, boxes):
        heights, widths = augmentation_utils.get_images_shape(
            images, dtype=tf.float32
        )
        x_scale = tf.cast(self.width / widths, dtype=self.compute_dtype)
        y_scale = tf.cast(self.height / heights, dtype=self.compute_dtype)
        x1s, y1s, x2s, y2s = tf.split(boxes, 4, axis=-1)
        outputs = tf.concat(
            [
                x1s * x_scale[:, tf.newaxis, :],
                y1s * y_scale[:, tf.newaxis, :],
                x2s * x_scale[:, tf.newaxis, :],
                y2s * y_scale[:, tf.newaxis, :],
            ],
            axis=-1,
        )
        return outputs

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "height": self.height,
                "width": self.width,
                "bounding_box_format": self.bounding_box_format,
                "bounding_box_min_area_ratio": self.bounding_box_min_area_ratio,  # noqa: E501
                "bounding_box_max_aspect_ratio": self.bounding_box_max_aspect_ratio,  # noqa: E501
                "seed": self.seed,
            }
        )
        return config



from functools import partial

import tensorflow as tf
from tensorflow import keras

from keras_aug import layers
from keras_aug.core import NormalFactorSampler
from keras_aug.core import SignedNormalFactorSampler
from keras_aug.datapoints import bounding_box
from keras_aug.datapoints import image as image_utils
from keras_aug.layers.base.vectorized_base_random_layer import (
    VectorizedBaseRandomLayer,
)
from keras_aug.utils import augmentation as augmentation_utils
from keras_aug.utils.augmentation import BOUNDING_BOXES
from keras_aug.utils.augmentation import IMAGES


@keras.utils.register_keras_serializable(package="keras_aug")
class RandAugment(VectorizedBaseRandomLayer):
    """RandAugment performs the Rand Augment operation on input images.

    RandAugment can be thought of as an all-in-one image augmentation layer. The
    policy implemented by RandAugment has been benchmarked extensively and is
    effective on a wide variety of datasets.

    The input images will be converted to the range [0, 255], performed
    RandAugment and then converted back to the original value range.

    For object detection tasks, you should set ``fill_mode="constant"`` and
    ``fill_value=128`` to avoid artifacts. Moreover, you can set
    ``use_geometry=False`` to turn off all geometric augmentations if the
    distortion of the bounding boxes is too large.

    Args:
        value_range (Sequence[int|float]): The range of values the incoming
            images will have. This is typically either ``[0, 1]`` or
            ``[0, 255]`` depending on how your preprocessing pipeline is set up.
        augmentations_per_image (int, optional): The number of layers to use in
            the rand augment policy. Defaults to ``2``.
        magnitude (float, optional): The shared magnitude across all
            augmentation operations. Represented as M in the paper. Usually best
            values are in the range ``[5, 10]``. Defaults to ``10``.
        magnitude_stddev (float, optional): The randomness of the severity as
            proposed by the authors of the timm library. Defaults to ``0``. When
            enabled, A gaussian noise with ``magnitude_stddev`` as sigma will be
            added to ``magnitude``.
        translation_multiplier (float, optional): The multiplier for applying
            translation. Defaults to ``150.0 / 331.0`` which is for ImageNet
            classification model. For CIFAR, it is set to ``10.0 / 32.0``.
            Usually best value is in the range ``[1.0 / 3.0, 1.0 / 2.0]``.
        use_geometry (bool, optional): whether to include geometric
            augmentations. This should be set to ``False`` when performing
            object detection. Defaults to ``True``.
        interpolation (str, optional): The interpolation mode. Supported values:
            ``"nearest", "bilinear"``. Defaults to `"nearest"`.
        fill_mode (str, optional): The fill mode. Supported values:
            ``"constant", "reflect", "wrap", "nearest"``. Defaults to
            ``"reflect"``.
        fill_value (int|float, optional): The value to be filled outside the
            boundaries when ``fill_mode="constant"``. Defaults to ``0``.
        exclude_ops (list(str), optional): Exclude selected operations.
            Defaults to ``None``.
        bounding_box_format (str, optional): The format of bounding
            boxes of input dataset. Refer
            https://github.com/james77777778/keras-aug/blob/main/keras_aug/datapoints/bounding_box/converter.py
            for more details on supported bounding box formats.
        seed (int|float, optional): The random seed. Defaults to ``None``.

    References:
        - `RandAugment <https://arxiv.org/abs/1909.13719>`_
        - `Tensorflow Model augment <https://github.com/tensorflow/models/blob/v2.12.0/official/vision/ops/augment.py>`_
        - `torchvision <https://github.com/pytorch/vision>`_
        - `KerasCV <https://github.com/keras-team/keras-cv>`_
    """  # noqa: E501

    def __init__(
        self,
        value_range,
        augmentations_per_image=2,
        magnitude=10,
        magnitude_stddev=0.0,
        translation_multiplier=150.0 / 331.0,
        use_geometry=True,
        interpolation="nearest",
        fill_mode="reflect",
        fill_value=0,
        exclude_ops=None,
        bounding_box_format=None,
        seed=None,
        **kwargs,
    ):
        super().__init__(seed=seed, **kwargs)
        self.value_range = value_range
        self.augmentations_per_image = augmentations_per_image
        self.magnitude = magnitude
        self.magnitude_stddev = magnitude_stddev
        self.translation_multiplier = translation_multiplier
        self.use_geometry = use_geometry
        self.interpolation = interpolation
        self.fill_mode = fill_mode
        self.fill_value = fill_value
        self.exclude_ops = exclude_ops
        self.bounding_box_format = bounding_box_format
        self.seed = seed

        self.aug_layers = self.get_standard_policy(
            magnitude,
            magnitude_stddev,
            translation_multiplier,
            use_geometry,
            exclude_ops,
            bounding_box_format,
            seed,
            **kwargs,
        )
        self.num_layers = len(self.aug_layers)

    def get_standard_policy(
        self,
        magnitude,
        magnitude_stddev,
        translation_multiplier,
        use_geometry,
        exclude_ops,
        bounding_box_format=None,
        seed=None,
        **kwargs,
    ):
        policy = create_rand_augment_policy(
            magnitude, magnitude_stddev, translation_multiplier, seed=seed
        )
        aug_layers = []
        if exclude_ops is not None:
            for op in exclude_ops:
                policy.pop(op)
        for key in policy.keys():
            if key == "identity":
                aug_layers.append(layers.Identity(**policy[key], **kwargs))
            elif key == "auto_contrast":
                aug_layers.append(
                    layers.AutoContrast(
                        **policy[key], value_range=(0, 255), seed=seed, **kwargs
                    )
                )
            elif key == "equalize":
                aug_layers.append(
                    layers.Equalize(
                        **policy[key], value_range=(0, 255), seed=seed, **kwargs
                    )
                )
            elif key == "posterize":
                aug_layers.append(
                    layers.RandomPosterize(
                        **policy[key], value_range=(0, 255), seed=seed, **kwargs
                    )
                )
            elif key == "solarize":
                aug_layers.append(
                    layers.RandomSolarize(
                        **policy[key], value_range=(0, 255), seed=seed, **kwargs
                    )
                )
            elif key == "color":
                aug_layers.append(
                    layers.RandomColorJitter(
                        **policy[key], value_range=(0, 255), seed=seed, **kwargs
                    )
                )
            elif key == "contrast":
                aug_layers.append(
                    layers.RandomColorJitter(
                        **policy[key], value_range=(0, 255), seed=seed, **kwargs
                    )
                )
            elif key == "brightness":
                aug_layers.append(
                    layers.RandomColorJitter(
                        **policy[key], value_range=(0, 255), seed=seed, **kwargs
                    )
                )
            elif key == "sharpness":
                aug_layers.append(
                    layers.RandomSharpness(
                        **policy[key], value_range=(0, 255), seed=seed, **kwargs
                    )
                )
            elif key == "rotate":
                if use_geometry:
                    aug_layers.append(
                        layers.RandomAffine(
                            **policy[key],
                            interpolation=self.interpolation,
                            fill_mode=self.fill_mode,
                            fill_value=self.fill_value,
                            bounding_box_format=bounding_box_format,
                            seed=seed,
                            **kwargs,
                        )
                    )
            elif key == "shear_x":
                if use_geometry:
                    aug_layers.append(
                        layers.RandomAffine(
                            **policy[key],
                            interpolation=self.interpolation,
                            fill_mode=self.fill_mode,
                            fill_value=self.fill_value,
                            bounding_box_format=bounding_box_format,
                            seed=seed,
                            **kwargs,
                        )
                    )
            elif key == "shear_y":
                if use_geometry:
                    aug_layers.append(
                        layers.RandomAffine(
                            **policy[key],
                            interpolation=self.interpolation,
                            fill_mode=self.fill_mode,
                            fill_value=self.fill_value,
                            bounding_box_format=bounding_box_format,
                            seed=seed,
                            **kwargs,
                        )
                    )
            elif key == "translate_x":
                if use_geometry:
                    aug_layers.append(
                        layers.RandomAffine(
                            **policy[key],
                            interpolation=self.interpolation,
                            fill_mode=self.fill_mode,
                            fill_value=self.fill_value,
                            bounding_box_format=bounding_box_format,
                            seed=seed,
                            **kwargs,
                        )
                    )
            elif key == "translate_y":
                if use_geometry:
                    aug_layers.append(
                        layers.RandomAffine(
                            **policy[key],
                            interpolation=self.interpolation,
                            fill_mode=self.fill_mode,
                            fill_value=self.fill_value,
                            bounding_box_format=bounding_box_format,
                            seed=seed,
                            **kwargs,
                        )
                    )
            else:
                raise ValueError(f"Not recognized policy key: {key}")
        return aug_layers

    def get_random_transformation_batch(self, batch_size):
        random_indices = self._random_generator.random_uniform(
            shape=(
                batch_size,
                self.augmentations_per_image,
            ),
            minval=0,
            maxval=self.num_layers,
            dtype=tf.int32,
        )
        return random_indices

    def _batch_augment(self, inputs):
        images = inputs.get(augmentation_utils.IMAGES, None)
        batch_size = tf.shape(images)[0]
        transformations = self.get_random_transformation_batch(batch_size)

        # images value_range transform to [0, 255]
        images = image_utils.transform_value_range(
            images, self.value_range, (0, 255), dtype=self.compute_dtype
        )
        inputs[IMAGES] = images

        bounding_boxes = inputs.get(BOUNDING_BOXES, None)
        # make bounding_boxes to dense first
        if bounding_boxes is not None:
            ori_bbox_info = bounding_box.validate_format(bounding_boxes)
            inputs[BOUNDING_BOXES] = bounding_box.to_dense(bounding_boxes)

        inputs_for_rand_augment_single_input = {
            "inputs": inputs,
            "transformations": transformations,
        }
        result = tf.map_fn(
            self.rand_augment_single_input,
            inputs_for_rand_augment_single_input,
            fn_output_signature=augmentation_utils.compute_signature(
                inputs, self.compute_dtype
            ),
        )

        bounding_boxes = result.get(BOUNDING_BOXES, None)
        if bounding_boxes is not None:
            if ori_bbox_info["ragged"]:
                bounding_boxes = bounding_box.to_ragged(bounding_boxes)
            else:
                bounding_boxes = bounding_box.to_dense(bounding_boxes)
            result[BOUNDING_BOXES] = bounding_boxes

        # recover value_range
        images = result.get(IMAGES, None)
        images = image_utils.transform_value_range(
            images, (0, 255), self.value_range, self.compute_dtype
        )
        result[IMAGES] = images
        return result

    def rand_augment_single_input(self, inputs):
        input = inputs.get("inputs")
        random_indices = inputs.get("transformations")

        # TODO:
        # figure out why tf will make tf.float32 instead of tf.float16
        # keras.mixed_precision.set_global_policy("mixed_float16")
        for i in range(self.augmentations_per_image):
            random_indice = random_indices[i]
            if BOUNDING_BOXES in input:
                tf.autograph.experimental.set_loop_options(
                    shape_invariants=[
                        (
                            input[BOUNDING_BOXES]["boxes"],
                            tf.TensorSpec([None, 4]),
                        ),
                        (
                            input[BOUNDING_BOXES]["classes"],
                            tf.TensorSpec([None]),
                        ),
                    ]
                )
            # construct branch_fns
            branch_fns = {}
            for idx, layer in enumerate(self.aug_layers):
                branch_fns[idx] = partial(layer, input)
            # augment
            input = tf.switch_case(random_indice, branch_fns=branch_fns)
            input = augmentation_utils.cast_to(input, self.compute_dtype)
        result = input
        if BOUNDING_BOXES in result:
            result[BOUNDING_BOXES] = bounding_box.to_ragged(
                result[BOUNDING_BOXES], dtype=self.compute_dtype
            )
        return result

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "value_range": self.value_range,
                "augmentations_per_image": self.augmentations_per_image,
                "magnitude": self.magnitude,
                "magnitude_stddev": self.magnitude_stddev,
                "translation_multiplier": self.translation_multiplier,
                "use_geometry": self.use_geometry,
                "interpolation": self.interpolation,
                "fill_mode": self.fill_mode,
                "fill_value": self.fill_value,
                "exclude_ops": self.exclude_ops,
                "bounding_box_format": self.bounding_box_format,
                "seed": self.seed,
            }
        )
        return config


def create_rand_augment_policy(
    magnitude, magnitude_stddev, translation_multiplier, seed
):
    """Create RandAugment Policy.

    References:
        https://github.com/pytorch/vision/blob/main/torchvision/transforms/v2/_auto_augment.py
    """  # noqa: E501
    max_magnitude = 30.0

    policy = {}
    policy["identity"] = {}
    policy["auto_contrast"] = {}
    policy["equalize"] = {}
    policy["rotate"] = {
        "rotation_factor": SignedNormalFactorSampler(
            mean=(magnitude / max_magnitude) * 30.0,
            stddev=magnitude_stddev * 30.0,
            min_value=0,
            max_value=30.0,
            seed=seed,
        ),
    }
    policy["posterize"] = {
        "factor": NormalFactorSampler(
            mean=8 - round(4 * (magnitude / max_magnitude)),  # must be int
            stddev=0,
            min_value=0,
            max_value=8,
            seed=seed,
        )
    }
    policy["solarize"] = {
        "threshold_factor": NormalFactorSampler(
            mean=255 - (magnitude / max_magnitude * 255),
            stddev=magnitude_stddev * 255,
            min_value=0,
            max_value=255,
            seed=seed,
        ),
        "addition_factor": 0,
    }
    policy["color"] = {
        "saturation_factor": NormalFactorSampler(
            mean=1.0 + magnitude / max_magnitude * 0.9,
            stddev=magnitude_stddev * 0.9,
            min_value=0,
            max_value=1.9,
            seed=seed,
        ),
    }
    policy["contrast"] = {
        "contrast_factor": NormalFactorSampler(
            mean=1.0 + magnitude / max_magnitude * 0.9,
            stddev=magnitude_stddev * 0.9,
            min_value=0,
            max_value=1.9,
            seed=seed,
        ),
    }
    policy["brightness"] = {
        "brightness_factor": NormalFactorSampler(
            mean=1.0 + magnitude / max_magnitude * 0.9,
            stddev=magnitude_stddev * 0.9,
            min_value=0,
            max_value=1.9,
            seed=seed,
        ),
    }
    policy["sharpness"] = {
        "factor": NormalFactorSampler(
            mean=1.0 + magnitude / max_magnitude * 0.9,
            stddev=magnitude_stddev * 0.9,
            min_value=0,
            max_value=1.9,
            seed=seed,
        )
    }
    policy["shear_x"] = {
        "shear_height_factor": 0,
        "shear_width_factor": SignedNormalFactorSampler(
            mean=magnitude / max_magnitude * 0.3,
            stddev=magnitude_stddev * 0.3,
            min_value=0,
            max_value=0.3,
            rate=0.5,
            seed=seed,
        ),
    }
    policy["shear_y"] = {
        "shear_height_factor": SignedNormalFactorSampler(
            mean=magnitude / max_magnitude * 0.3,
            stddev=magnitude_stddev * 0.3,
            min_value=0,
            max_value=0.3,
            rate=0.5,
            seed=seed,
        ),
        "shear_width_factor": 0,
    }
    policy["translate_x"] = {
        "translation_height_factor": 0,
        "translation_width_factor": SignedNormalFactorSampler(
            mean=magnitude / max_magnitude * translation_multiplier,
            stddev=magnitude_stddev * translation_multiplier,
            min_value=0,
            max_value=translation_multiplier,
            rate=0.5,
            seed=seed,
        ),
    }
    policy["translate_y"] = {
        "translation_height_factor": SignedNormalFactorSampler(
            mean=magnitude / max_magnitude * translation_multiplier,
            stddev=magnitude_stddev * translation_multiplier,
            min_value=0,
            max_value=translation_multiplier,
            rate=0.5,
            seed=seed,
        ),
        "translation_width_factor": 0,
    }
    return policy



import tensorflow as tf

from keras_aug import layers


class RandomGammaTest(tf.test.TestCase):
    regular_args = {
        "value_range": (0, 255),
        "factor": (1.0 - 0.25, 1.0 + 0.25),
    }
    no_aug_args = {
        "value_range": (0, 255),
        "factor": (1.0, 1.0),
    }

    def test_no_adjustment(self):
        image_shape = (4, 8, 8, 3)
        image = tf.random.uniform(shape=image_shape) * 255.0
        layer = layers.RandomGamma(**self.no_aug_args)

        output = layer(image)

        self.assertAllClose(image, output, atol=1e-5, rtol=1e-5)

    def test_adjustment_for_non_rgb_value_range(self):
        image_shape = (4, 8, 8, 3)
        # Value range (0, 100)
        image = tf.random.uniform(shape=image_shape) * 100.0

        layer = layers.RandomGamma(**self.no_aug_args)
        output = layer(image)
        self.assertAllClose(image, output, atol=1e-5, rtol=1e-5)

        layer = layers.RandomGamma(**self.regular_args)
        output = layer(image)
        self.assertNotAllClose(image, output)

    def test_sqrt_adjustment(self):
        image_shape = (4, 8, 8, 3)
        image = tf.random.uniform(shape=image_shape)
        args = self.no_aug_args.copy()
        args.update({"value_range": (0, 1), "factor": (0.5, 0.5)})
        layer = layers.RandomGamma(**args)

        output = layer(image)

        self.assertAllClose(tf.sqrt(image), output, atol=1e-5, rtol=1e-5)



import keras
import tensorflow as tf

from keras_aug.core.factor_sampler.factor_sampler import FactorSampler
from keras_aug.core.keras_random_generator import KerasRandomGenerator


@keras.utils.register_keras_serializable(package="keras_aug")
class SignedNormalFactorSampler(FactorSampler):
    """SignedNormalFactorSampler samples factors from a normal distribution and
    randomly inverts the sampled factors.

    This is useful in cases where a user wants to always ensure that an
    augmentation layer performs augmentations of the same strength.

    Args:
        mean (float): The mean value for the distribution.
        stddev (float): The standard deviation of the distribution.
        min_value (float): The values below ``min_value`` are clipped to
            ``min_value``.
        max_value (float): The values above ``max_value`` are clipped to
            ``max_value``.
        rate (float, optional): The rate to invert factors. Defaults to ``0.5``.
        seed (int|float, optional): The random seed. Defaults to ``None``.

    References:
        - `KerasCV <https://github.com/keras-team/keras-cv>`_
    """

    def __init__(self, mean, stddev, min_value, max_value, rate=0.5, seed=None):
        self.mean = mean
        self.stddev = stddev
        self.min_value = min_value
        self.max_value = max_value
        self.rate = rate
        self.seed = seed
        self.rng = KerasRandomGenerator(seed=seed, rng_type="stateless")

    def __call__(self, shape=(), dtype="float32"):
        if self.stddev != 0:
            factors = self.rng.random_normal(
                shape=shape,
                mean=self.mean,
                stddev=self.stddev,
                dtype=dtype,
            )
        else:
            # if self.stddev == 0, degrade to SignedConstantFactorSampler
            factors = tf.ones(shape=shape, dtype=dtype) * self.mean
        factors = tf.clip_by_value(factors, self.min_value, self.max_value)
        negates = self.rng.random_uniform(
            shape=shape, minval=0, maxval=1, dtype=tf.float32
        )
        negates = tf.cast(tf.where(negates > self.rate, -1.0, 1.0), dtype=dtype)
        return factors * negates

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "mean": self.mean,
                "stddev": self.stddev,
                "min_value": self.min_value,
                "max_value": self.max_value,
                "rate": self.rate,
                "seed": self.seed,
            }
        )
        return config



import tensorflow as tf
from tensorflow import keras

from keras_aug.layers.base.vectorized_base_random_layer import (
    VectorizedBaseRandomLayer,
)
from keras_aug.utils import augmentation as augmentation_utils


@keras.utils.register_keras_serializable(package="keras_aug")
class RandomGaussianBlur(VectorizedBaseRandomLayer):
    """Applies a Gaussian Blur with random strength to an image.

    Args:
        kernel_size (int|Sequence[int]): The x and y dimensions for the kernel
            used.
        factor (float|Sequence[float]|keras_aug.FactorSampler): The range of the
            factor that controls the extent to which the image is blurred.
            When represented as a single float, the factor will be picked
            between ``[0.0, 0.0 + upper]``. Mathematically, ``factor``
            represents the sigma value in a gaussian blur. ``0.0`` makes this
            layer perform a no-op operation. High values make the blur
            stronger.
        seed (int|float, optional): The random seed. Defaults to ``None``.

    References:
        - `KerasCV <https://github.com/keras-team/keras-cv>`_
    """

    def __init__(self, kernel_size, factor, seed=None, **kwargs):
        super().__init__(**kwargs)
        if isinstance(kernel_size, (tuple, list)):
            self.x = kernel_size[0]
            self.y = kernel_size[1]
        else:
            if isinstance(kernel_size, int):
                self.x = self.y = kernel_size
            else:
                raise ValueError(
                    "`kernel_size` must be list, tuple or integer "
                    ", got {} ".format(type(self.kernel_size))
                )
        self.kernel_size = kernel_size
        if isinstance(factor, (int, float)):
            factor = (0.0, factor)
        self.factor = augmentation_utils.parse_factor(
            factor, min_value=0.0, max_value=None, seed=seed
        )

    def get_random_transformation_batch(self, batch_size, **kwargs):
        factors = self.factor(shape=(batch_size, 1))
        blur_vs = self.get_kernel(factors, self.y)
        blur_hs = self.get_kernel(factors, self.x)
        blur_vs = tf.reshape(blur_vs, [batch_size, self.y, 1, 1, 1])
        blur_hs = tf.reshape(blur_hs, [batch_size, 1, self.x, 1, 1])
        return {"blur_vs": blur_vs, "blur_hs": blur_hs}

    def augment_ragged_image(self, image, transformation, **kwargs):
        images = tf.expand_dims(image, axis=0)
        transformations = augmentation_utils.expand_dict_dims(
            transformation, axis=0
        )
        images = self.augment_images(
            images=images, transformations=transformations, **kwargs
        )
        return tf.squeeze(images, axis=0)

    def augment_images(self, images, transformations, **kwargs):
        images = tf.cast(images, dtype=self.compute_dtype)
        num_channels = tf.shape(images)[-1]
        blur_vs = transformations["blur_vs"]
        blur_hs = transformations["blur_hs"]
        blur_vs = tf.tile(blur_vs, [1, 1, 1, num_channels, 1])
        blur_vs = tf.cast(blur_vs, dtype=self.compute_dtype)
        blur_hs = tf.tile(blur_hs, [1, 1, 1, num_channels, 1])
        blur_hs = tf.cast(blur_hs, dtype=self.compute_dtype)
        inputs_for_gaussian_blur_single_image = {
            "images": images,
            "blur_vs": blur_vs,
            "blur_hs": blur_hs,
        }
        images = tf.vectorized_map(
            self.gaussian_blur_single_image,
            inputs_for_gaussian_blur_single_image,
        )
        return images

    def augment_labels(self, labels, transformations, **kwargs):
        return labels

    def augment_bounding_boxes(self, bounding_boxes, transformations, **kwargs):
        return bounding_boxes

    def augment_segmentation_masks(
        self, segmentation_masks, transformations, **kwargs
    ):
        return segmentation_masks

    def augment_keypoints(self, keypoints, transformations, **kwargs):
        return keypoints

    def get_kernel(self, factor, filter_size):
        # We are running this in float32, regardless of layer's
        # self.compute_dtype. Calculating blur_filter in lower precision will
        # corrupt the final results.
        x = tf.cast(
            tf.range(-filter_size // 2 + 1, filter_size // 2 + 1),
            dtype=tf.float32,
        )
        x = tf.reshape(x, shape=(1, -1))
        blur_filter = tf.exp(
            -tf.pow(x, 2.0)
            / (2.0 * tf.pow(tf.cast(factor, dtype=tf.float32), 2.0))
        )
        blur_filter /= tf.reduce_sum(blur_filter, axis=-1, keepdims=True)
        return blur_filter

    def gaussian_blur_single_image(self, inputs):
        image = tf.expand_dims(inputs.get("images", None), axis=0)
        blur_v = inputs.get("blur_vs", None)
        blur_h = inputs.get("blur_hs", None)
        image = tf.nn.depthwise_conv2d(
            image, blur_h, strides=[1, 1, 1, 1], padding="SAME"
        )
        image = tf.nn.depthwise_conv2d(
            image, blur_v, strides=[1, 1, 1, 1], padding="SAME"
        )
        return tf.squeeze(image, axis=0)

    def get_config(self):
        config = super().get_config()
        config.update({"factor": self.factor, "kernel_size": self.kernel_size})
        return config



import unittest

import numpy as np
import tensorflow as tf
from absl.testing import parameterized
from tensorflow import keras

from keras_aug import layers


class RandomCropTest(tf.test.TestCase, parameterized.TestCase):
    def test_input_smaller_than_crop_box(self):
        np.random.seed(1337)
        height, width = 10, 8
        inp = np.random.random((12, 3, 3, 3))
        layer = layers.RandomCrop(height, width)
        actual_output = layer(inp)
        # In this case, output should equal resizing with crop_to_aspect
        # ratio.
        resizing_layer = keras.layers.Resizing(height, width)
        expected_output = resizing_layer(inp)
        self.assertAllEqual(expected_output, actual_output)

    def test_training_with_mock(self):
        np.random.seed(1337)
        batch_size = 12
        height, width = 3, 4
        height_offset = np.random.randint(low=0, high=3)
        width_offset = np.random.randint(low=0, high=5)
        # manually compute transformations which shift height_offset and
        # width_offset respectively
        tops = tf.ones((batch_size, 1)) * (height_offset / (5 - height))
        lefts = tf.ones((batch_size, 1)) * (width_offset / (8 - width))
        transformations = {"crop_tops": tops, "crop_lefts": lefts}
        layer = layers.RandomCrop(height, width)
        with unittest.mock.patch.object(
            layer,
            "get_random_transformation_batch",
            return_value=transformations,
        ):
            inp = np.random.random((12, 5, 8, 3))
            actual_output = layer(inp)
            expected_output = inp[
                :,
                height_offset : (height_offset + height),
                width_offset : (width_offset + width),
                :,
            ]
            self.assertAllClose(expected_output, actual_output)

    def test_random_crop_full(self):
        np.random.seed(1337)
        height, width = 8, 16
        inp = np.random.random((12, 8, 16, 3))
        layer = layers.RandomCrop(height, width)
        actual_output = layer(inp)
        self.assertAllClose(inp, actual_output)

    def test_unbatched_image(self):
        np.random.seed(1337)
        inp = np.random.random((16, 16, 3))
        # manually compute transformations which shift 2 pixels
        mock_offset = tf.ones(shape=(1, 1), dtype=tf.float32) * 0.25
        layer = layers.RandomCrop(8, 8)
        with unittest.mock.patch.object(
            layer._random_generator,
            "random_uniform",
            return_value=mock_offset,
        ):
            actual_output = layer(inp)
            self.assertAllClose(inp[2:10, 2:10, :], actual_output)

    def test_batched_input(self):
        np.random.seed(1337)
        inp = np.random.random((20, 16, 16, 3))
        # manually compute transformations which shift 2 pixels
        mock_offset = tf.ones(shape=(20, 1), dtype=tf.float32) * 2 / (16 - 8)
        layer = layers.RandomCrop(8, 8)
        with unittest.mock.patch.object(
            layer._random_generator,
            "random_uniform",
            return_value=mock_offset,
        ):
            actual_output = layer(inp)
            self.assertAllClose(inp[:, 2:10, 2:10, :], actual_output)

    def test_augment_bounding_boxes_crop(self):
        orig_height, orig_width = 512, 512
        height, width = 100, 200
        input_image = np.random.random((orig_height, orig_width, 3)).astype(
            np.float32
        )
        bboxes = {
            "boxes": tf.convert_to_tensor([[200, 200, 400, 400]]),
            "classes": tf.convert_to_tensor([1]),
        }
        input = {"images": input_image, "bounding_boxes": bboxes}
        # for top = 300 and left = 305
        height_offset = 300
        width_offset = 305
        tops = tf.ones((1, 1)) * (height_offset / (orig_height - height))
        lefts = tf.ones((1, 1)) * (width_offset / (orig_width - width))
        transformations = {"crop_tops": tops, "crop_lefts": lefts}
        layer = layers.RandomCrop(
            height=height, width=width, bounding_box_format="xyxy"
        )
        with unittest.mock.patch.object(
            layer,
            "get_random_transformation_batch",
            return_value=transformations,
        ):
            output = layer(input)
            expected_output = np.asarray(
                [[0.0, 0.0, 95.0, 100.0]],
            )
        self.assertAllClose(expected_output, output["bounding_boxes"]["boxes"])

    def test_augment_bounding_boxes_resize(self):
        input_image = np.random.random((256, 256, 3)).astype(np.float32)
        bboxes = {
            "boxes": tf.convert_to_tensor([[100, 100, 200, 200]]),
            "classes": tf.convert_to_tensor([1]),
        }
        input = {"images": input_image, "bounding_boxes": bboxes}
        layer = layers.RandomCrop(
            height=512, width=512, bounding_box_format="xyxy"
        )
        output = layer(input)
        expected_output = np.asarray(
            [[200.0, 200.0, 400.0, 400.0]],
        )
        self.assertAllClose(expected_output, output["bounding_boxes"]["boxes"])

    def test_random_crop_on_batched_ragged_images_and_bounding_boxes(self):
        images = tf.ragged.constant(
            [np.ones((8, 8, 3)), np.ones((4, 8, 3))], dtype="float32"
        )
        boxes = {
            "boxes": tf.ragged.stack(
                [
                    tf.ones((3, 4), dtype=tf.float32),
                    tf.ones((3, 4), dtype=tf.float32),
                ],
            ),
            "classes": tf.ragged.stack(
                [
                    tf.ones((3,), dtype=tf.float32),
                    tf.ones((3,), dtype=tf.float32),
                ],
            ),
        }
        inputs = {"images": images, "bounding_boxes": boxes}
        layer = layers.RandomCrop(height=2, width=2, bounding_box_format="xyxy")

        results = layer(inputs)

        self.assertTrue(isinstance(results["images"], tf.Tensor))
        self.assertTrue(
            isinstance(results["bounding_boxes"]["boxes"], tf.RaggedTensor)
        )
        self.assertTrue(
            isinstance(results["bounding_boxes"]["classes"], tf.RaggedTensor)
        )

    def test_dense_segmentation_masks(self):
        images = tf.random.uniform((2, 10, 10, 3))
        segmentation_masks = tf.random.uniform(
            (2, 10, 10, 1), minval=0, maxval=10, dtype=tf.int32
        )
        layer = layers.RandomCrop(height=2, width=2)

        result = layer(
            {"images": images, "segmentation_masks": segmentation_masks}
        )

        self.assertTrue(isinstance(result["segmentation_masks"], tf.Tensor))
        self.assertEqual(result["segmentation_masks"].shape[1:3], (2, 2))
        self.assertAllInSet(result["segmentation_masks"], tf.range(0, 10))

    def test_ragged_segmentation_masks(self):
        images = tf.ragged.stack(
            [
                tf.random.uniform((8, 8, 3), dtype=tf.float32),
                tf.random.uniform((16, 8, 3), dtype=tf.float32),
            ]
        )
        segmentation_masks = tf.ragged.stack(
            [
                tf.random.uniform((8, 8, 1), maxval=10, dtype=tf.int32),
                tf.random.uniform((16, 8, 1), maxval=10, dtype=tf.int32),
            ]
        )
        segmentation_masks = tf.cast(segmentation_masks, dtype=tf.float32)
        layer = layers.RandomCrop(height=2, width=2)

        result = layer(
            {"images": images, "segmentation_masks": segmentation_masks}
        )

        self.assertTrue(isinstance(result["segmentation_masks"], tf.Tensor))
        self.assertEqual(result["segmentation_masks"].shape[1:3], (2, 2))
        self.assertAllInSet(result["segmentation_masks"], tf.range(0, 10))

    def test_ragged_input_with_graph_mode(self):
        images = tf.ragged.stack(
            [
                tf.random.uniform((8, 8, 3), dtype=tf.float32),
                tf.random.uniform((16, 8, 3), dtype=tf.float32),
            ]
        )
        segmentation_masks = tf.ragged.stack(
            [
                tf.random.uniform((8, 8, 1), maxval=10, dtype=tf.int32),
                tf.random.uniform((16, 8, 1), maxval=10, dtype=tf.int32),
            ]
        )
        layer = layers.RandomCrop(height=8, width=8)

        @tf.function
        def fn(inputs):
            outputs = layer(inputs)
            image_shape = outputs["images"].shape
            segmentation_mask_shape = outputs["segmentation_masks"].shape
            assert image_shape == (2, 8, 8, 3)
            assert segmentation_mask_shape == (2, 8, 8, 1)

        fn({"images": images, "segmentation_masks": segmentation_masks})



"""
Most of these codes come from KerasCV.
"""
import tensorflow as tf
from keras import backend


def _unwrap_value_range(value_range, dtype=tf.float32):
    min_value, max_value = value_range
    min_value = tf.cast(min_value, dtype=dtype)
    max_value = tf.cast(max_value, dtype=dtype)
    return min_value, max_value


def transform_value_range(
    images, original_range, target_range, dtype=tf.float32
):
    """transforms values in input tensor from original_range to target_range.
    This function is intended to be used in preprocessing layers that
    rely upon color values. This allows us to assume internally that
    the input tensor is always in the range [0, 255].

    Args:
        images (tf.Tensor): the set of images to transform to the target range.
        original_range (Tuple[int|float, int|float]): The value range to
            transform from.
        target_range (Tuple[int|float, int|float]): The value range to
            transform to.
        dtype (tf.DType, optional): The dtype to compute the conversion with.
            Defaults to `tf.float32`.

    Returns:
        a new Tensor with values in the target range.

    References:
        - `KerasCV <https://github.com/keras-team/keras-cv>`_
    """
    if (
        original_range[0] == target_range[0]
        and original_range[1] == target_range[1]
    ):
        return images

    images = tf.cast(images, dtype=dtype)
    original_min_value, original_max_value = _unwrap_value_range(
        original_range, dtype=dtype
    )
    target_min_value, target_max_value = _unwrap_value_range(
        target_range, dtype=dtype
    )

    # images in the [0, 1] scale
    images = (images - original_min_value) / (
        original_max_value - original_min_value
    )

    scale_factor = target_max_value - target_min_value
    return (images * scale_factor) + target_min_value


def blend(images_1, images_2, factors, value_range=None):
    """Blend image1 and image2 using 'factors'. Can be batched inputs.

    Factor can be above ``0.0``.  A value of ``0.0`` means only image1 is used.
    A value of ``1.0`` means only image2 is used.  A value between ``0.0`` and
    ``1.0`` means we linearly interpolate the pixel values between the two
    images.  A value greater than ``1.0`` "extrapolates" the difference
    between the two pixel values. If ``value_range`` is set, the results will be
    clipped into ``value_range``

    Args:
        image1 (tf.Tensor): First image(s).
        image2 (tf.Tensor): Second image(s).
        factor (float|tf.Tensor): The blend factor(s).
        value_range (Sequence[int|float], optional): The value range of the
            results. Defaults to ``None``.

    References:
        - `KerasCV <https://github.com/keras-team/keras-cv>`_
    """
    results = images_1 + factors * (images_2 - images_1)
    if value_range is not None:
        results = tf.clip_by_value(results, value_range[0], value_range[1])
    return results


def rgb_to_grayscale(images):
    """Converts images from RGB to Grayscale.

    Compared to ``tf.image.rgb_to_grayscale``, this function replaces
    ``tf.tensordot`` with ``tf.math.multiply`` and ``tf.math.add`` to reduce
    memory usage.

    Args:
        images (tf.Tensor): The RGB tensor to convert. The last dimension must
            have size 3 and should contain RGB values.

    References:
        - `torchvision <https://github.com/pytorch/vision>`_
    """
    return (
        images[..., 0:1] * 0.2989
        + images[..., 1:2] * 0.587
        + images[..., 2:3] * 0.114
    )


def transform(
    images,
    transforms,
    fill_mode="reflect",
    fill_value=0.0,
    interpolation="bilinear",
    output_shape=None,
    name=None,
):
    """Applies the given transform(s) to the image(s).

    Args:
      images: A tensor of shape
        `(num_images, num_rows, num_columns, num_channels)` (NHWC). The rank
        must be statically known (the shape is not `TensorShape(None)`).
      transforms: Projective transform matrix/matrices. A vector of length 8 or
        tensor of size N x 8. If one row of transforms is
        [a0, a1, a2, b0, b1, b2, c0, c1], then it maps the *output* point
        `(x, y)` to a transformed *input* point
        `(x', y') = ((a0 x + a1 y + a2) / k, (b0 x + b1 y + b2) / k)`, where
        `k = c0 x + c1 y + 1`. The transforms are *inverted* compared to the
        transform mapping input points to output points. Note that gradients are
        not backpropagated into transformation parameters.
      fill_mode: Points outside the boundaries of the input are filled according
        to the given mode (one of `{"constant", "reflect", "wrap", "nearest"}`).
      fill_value: a float represents the value to be filled outside the
        boundaries when `fill_mode="constant"`.
      interpolation: Interpolation mode. Supported values: `"nearest"`,
        `"bilinear"`.
      output_shape: Output dimension after the transform, `[height, width]`.
        If `None`, output is the same size as input image.
      name: The name of the op.

    Fill mode behavior for each valid value is as follows:

    - reflect (d c b a | a b c d | d c b a)
    The input is extended by reflecting about the edge of the last pixel.

    - constant (k k k k | a b c d | k k k k)
    The input is extended by filling all
    values beyond the edge with the same constant value k = 0.

    - wrap (a b c d | a b c d | a b c d)
    The input is extended by wrapping around to the opposite edge.

    - nearest (a a a a | a b c d | d d d d)
    The input is extended by the nearest pixel.

    Input shape:
      4D tensor with shape: `(samples, height, width, channels)`,
        in `"channels_last"` format.

    Output shape:
      4D tensor with shape: `(samples, height, width, channels)`,
        in `"channels_last"` format.

    Returns:
      Image(s) with the same type and shape as `images`, with the given
      transform(s) applied. Transformed coordinates outside of the input image
      will be filled with zeros.

    Raises:
      TypeError: If `image` is an invalid type.
      ValueError: If output shape is not 1-D int32 Tensor.

    References:
        - `KerasCV <https://github.com/keras-team/keras-cv>`_
    """
    with backend.name_scope(name or "transform"):
        if output_shape is None:
            output_shape = tf.shape(images)[1:3]
            if not tf.executing_eagerly():
                output_shape_value = tf.get_static_value(output_shape)
                if output_shape_value is not None:
                    output_shape = output_shape_value

        output_shape = tf.convert_to_tensor(
            output_shape, tf.int32, name="output_shape"
        )

        if not output_shape.get_shape().is_compatible_with([2]):
            raise ValueError(
                "output_shape must be a 1-D Tensor of 2 elements: "
                "new_height, new_width, instead got "
                "{}".format(output_shape)
            )

        fill_value = tf.convert_to_tensor(
            fill_value, tf.float32, name="fill_value"
        )

        return tf.raw_ops.ImageProjectiveTransformV3(
            images=images,
            output_shape=output_shape,
            fill_value=fill_value,
            transforms=transforms,
            fill_mode=fill_mode.upper(),
            interpolation=interpolation.upper(),
        )



import tensorflow as tf

try:
    import tensorflow_datasets as tfds
except ImportError:
    tfds = None
try:
    import keras_cv
except ImportError:
    keras_cv = None

from keras_aug.datapoints import bounding_box
from keras_aug.datapoints import image as image_utils
from keras_aug.utils.conditional_imports import assert_kerascv_installed
from keras_aug.utils.conditional_imports import assert_tfds_installed


def load_voc_dataset(
    bounding_box_format,
    name="voc/2007",
    batch_size=9,
):
    assert_tfds_installed("load_voc_dataset")

    def preprocess_voc(inputs, format=None):
        image = inputs["image"]
        image = tf.cast(image, tf.float32)
        if format is not None:
            boxes = inputs["objects"]["bbox"]
            boxes = bounding_box.convert_format(
                boxes,
                images=image,
                source="rel_yxyx",
                target=format,
            )
            classes = tf.cast(inputs["objects"]["label"], tf.float32)
            bounding_boxes = {"classes": classes, "boxes": boxes}
            return {"images": image, "bounding_boxes": bounding_boxes}
        else:
            return {"images": image}

    dataset = tfds.load(name, split=tfds.Split.TRAIN, shuffle_files=False)
    dataset = dataset.map(
        lambda x: preprocess_voc(x, format=bounding_box_format),
        num_parallel_calls=tf.data.AUTOTUNE,
    )
    if tf.__version__ >= "2.12.0":
        dataset = dataset.ragged_batch(batch_size, drop_remainder=True)
    else:
        dataset = dataset.apply(
            tf.data.experimental.dense_to_ragged_batch(
                batch_size, drop_remainder=True
            )
        )
    return dataset


def load_oxford_dataset(
    name="oxford_flowers102",
    batch_size=9,
    img_size=(224, 224),
    as_supervised=True,
):
    assert_tfds_installed("oxford_dataset")

    def preprocess_oxford(image, label, img_size=(224, 224), num_classes=10):
        image = tf.image.resize(image, img_size)
        label = tf.one_hot(label, num_classes)
        return {"images": image, "labels": label}

    data, ds_info = tfds.load(
        name, as_supervised=as_supervised, with_info=True, shuffle_files=False
    )
    dataset = data["train"]
    num_classes = ds_info.features["label"].num_classes
    dataset = dataset.map(
        lambda x, y: preprocess_oxford(
            x, y, img_size=img_size, num_classes=num_classes
        ),
        num_parallel_calls=tf.data.AUTOTUNE,
    )
    dataset = dataset.batch(batch_size)
    return dataset


def load_oxford_iiit_pet_dataset(
    name="oxford_iiit_pet:3.*.*",
    batch_size=9,
    img_size=(224, 224),
):
    assert_tfds_installed("oxford_iiit_pet_dataset")

    def preprocess_oxford_iiit_pet(data):
        img_size = (224, 224)

        input_image = tf.image.resize(data["image"], img_size)
        input_mask = tf.image.resize(
            data["segmentation_mask"],
            img_size,
            method="nearest",
        )
        input_image = tf.image.convert_image_dtype(input_image, tf.float32)
        input_mask -= 1
        input_mask = tf.cast(input_mask, tf.float32)
        return {"images": input_image, "segmentation_masks": input_mask}

    data = tfds.load(name)
    dataset = data["train"]
    dataset = dataset.map(
        preprocess_oxford_iiit_pet,
        num_parallel_calls=tf.data.AUTOTUNE,
    )
    dataset = dataset.batch(batch_size)
    return dataset


def visualize_data(
    data, value_range=(0, 255), bounding_box_format=None, output_path=None
):
    assert_kerascv_installed("visualize_data")

    data = next(iter(data))
    images = data["images"]
    if isinstance(images, tf.RaggedTensor):
        images = images.to_tensor(0)
    if "labels" in data:
        mask = tf.greater(data["labels"][0], 0)
        non_zero_array = tf.boolean_mask(data["labels"][0], mask)
        print(f"Nonzero labels of the first image:\n{non_zero_array}")
    if bounding_box_format is not None:
        bounding_boxes = data["bounding_boxes"]
        keras_cv.visualization.plot_bounding_box_gallery(
            images,
            value_range=value_range,
            bounding_box_format=bounding_box_format,
            y_true=bounding_boxes,
            path=output_path,
            dpi=100,
        )
    else:
        keras_cv.visualization.plot_image_gallery(
            images,
            value_range=value_range,
            path=output_path,
            dpi=100,
        )


def visualize_data_single(
    data, value_range=(0, 255), bounding_box_format=None, output_path=None
):
    assert_kerascv_installed("visualize_data")

    data = next(iter(data))
    images = data["images"]
    images = images[0:1, ...]
    if isinstance(images, tf.RaggedTensor):
        images = images.to_tensor(0)
    if "labels" in data:
        mask = tf.greater(data["labels"][0], 0)
        non_zero_array = tf.boolean_mask(data["labels"][0], mask)
        print(f"Nonzero labels of the first image:\n{non_zero_array}")
    if bounding_box_format is not None:
        bounding_boxes = data["bounding_boxes"]
        bounding_boxes = {
            "boxes": bounding_boxes["boxes"][0:1, ...],
            "classes": bounding_boxes["classes"][0:1, ...],
        }
        keras_cv.visualization.plot_bounding_box_gallery(
            images,
            value_range=value_range,
            bounding_box_format=bounding_box_format,
            y_true=bounding_boxes,
            path=output_path,
            line_thickness=1,
            rows=1,
            cols=1,
            dpi=300,
        )
    else:
        keras_cv.visualization.plot_image_gallery(
            images,
            value_range=value_range,
            path=output_path,
            rows=1,
            cols=1,
            dpi=300,
        )


def visualize_segmentation_masks(
    data,
    image_value_range=(0, 255),
    mask_value_range=None,
    output_path=None,
):
    assert_kerascv_installed("visualize_data")

    data = next(iter(data))
    images = data["images"]
    masks = data["segmentation_masks"]
    masks = image_utils.transform_value_range(masks, mask_value_range, (0, 255))
    masks = tf.concat([masks, masks, masks], axis=-1)
    display_images = tf.concat([images, masks], axis=2)  # B, H, W, C

    keras_cv.visualization.plot_image_gallery(
        display_images,
        value_range=image_value_range,
        path=output_path,
        dpi=100,
    )



import tensorflow as tf

from keras_aug import layers
from keras_aug.datapoints import bounding_box


class MixUpTest(tf.test.TestCase):
    num_classes = 10

    def test_return_shapes(self):
        xs = tf.ones((2, 4, 4, 3))
        # randomly sample labels
        ys_labels = tf.random.categorical(tf.math.log([[0.5, 0.5]]), 2)
        ys_labels = tf.squeeze(ys_labels)
        ys_labels = tf.one_hot(ys_labels, self.num_classes)
        # randomly sample bounding boxes
        ys_bounding_boxes = {
            "boxes": tf.random.uniform((2, 3, 4), 0, 1),
            "classes": tf.random.uniform((2, 3), 0, 1),
        }
        layer = layers.MixUp()

        # mixup on labels
        outputs = layer(
            {
                "images": xs,
                "labels": ys_labels,
                "bounding_boxes": ys_bounding_boxes,
            }
        )
        xs, ys_labels, ys_bounding_boxes = (
            outputs["images"],
            outputs["labels"],
            outputs["bounding_boxes"],
        )
        ys_bounding_boxes = bounding_box.to_dense(ys_bounding_boxes)

        self.assertEqual(xs.shape, [2, 4, 4, 3])
        self.assertEqual(ys_labels.shape, [2, 10])
        self.assertEqual(ys_bounding_boxes["boxes"].shape, [2, 6, 4])
        self.assertEqual(ys_bounding_boxes["classes"].shape, [2, 6])

    def test_mix_up_call_results(self):
        xs = tf.cast(
            tf.stack(
                [2 * tf.ones((4, 4, 3)), tf.ones((4, 4, 3))],
                axis=0,
            ),
            tf.float32,
        )
        ys = tf.one_hot(tf.constant([0, 1]), 2)

        layer = layers.MixUp()
        outputs = layer({"images": xs, "labels": ys})
        xs, ys = outputs["images"], outputs["labels"]

        # None of the individual values should still be close to 1 or 0
        self.assertNotAllClose(xs, 1.0)
        self.assertNotAllClose(xs, 2.0)

        # No labels should still be close to their originals
        self.assertNotAllClose(ys, 1.0)
        self.assertNotAllClose(ys, 0.0)

    def test_image_input_only(self):
        xs = tf.cast(
            tf.stack([2 * tf.ones((4, 4, 1)), tf.ones((4, 4, 1))], axis=0),
            tf.float32,
        )
        layer = layers.MixUp()
        with self.assertRaisesRegexp(
            ValueError, "expects inputs in a dictionary"
        ):
            _ = layer(xs)

    def test_single_image_input(self):
        xs = tf.ones((4, 4, 3))
        ys = tf.one_hot(tf.constant([1]), 2)
        inputs = {"images": xs, "labels": ys}
        layer = layers.MixUp()
        with self.assertRaisesRegexp(
            ValueError, "MixUp received a single image to `call`"
        ):
            _ = layer(inputs)

    def test_image_input(self):
        xs = tf.ones((2, 4, 4, 3))
        layer = layers.MixUp()
        with self.assertRaisesRegexp(
            ValueError, "MixUp expects inputs in a dictionary with format"
        ):
            _ = layer(xs)



import tensorflow as tf

from keras_aug import layers

num_classes = 10


class CutMixTest(tf.test.TestCase):
    def test_return_shapes(self):
        xs = tf.ones((2, 40, 40, 3))
        # randomly sample labels
        ys = tf.random.categorical(tf.math.log([[0.5, 0.5]]), 2)
        ys = tf.squeeze(ys)
        ys = tf.one_hot(ys, num_classes)
        layer = layers.CutMix(seed=2023)
        outputs = layer({"images": xs, "labels": ys})

        xs, ys = outputs["images"], outputs["labels"]

        self.assertEqual(xs.shape, [2, 40, 40, 3])
        self.assertEqual(ys.shape, [2, 10])

    def test_cut_mix_call_results(self):
        xs = tf.cast(
            tf.stack(
                [2 * tf.ones((40, 40, 3)), tf.ones((40, 40, 3))],
                axis=0,
            ),
            tf.float32,
        )
        ys = tf.one_hot(tf.constant([0, 1]), 2)
        layer = layers.CutMix(seed=2024)

        outputs = layer({"images": xs, "labels": ys})
        xs, ys = outputs["images"], outputs["labels"]

        # At least some pixels should be replaced in the CutMix operation
        self.assertTrue(tf.math.reduce_any(xs[0] == 1.0))
        self.assertTrue(tf.math.reduce_any(xs[0] == 2.0))
        self.assertTrue(tf.math.reduce_any(xs[1] == 1.0))
        self.assertTrue(tf.math.reduce_any(xs[1] == 2.0))
        # No labels should still be close to their original values
        self.assertNotAllClose(ys, 1.0)
        self.assertNotAllClose(ys, 0.0)

    def test_cut_mix_call_results_one_channel(self):
        xs = tf.cast(
            tf.stack(
                [2 * tf.ones((40, 40, 1)), tf.ones((40, 40, 1))],
                axis=0,
            ),
            tf.float32,
        )
        ys = tf.one_hot(tf.constant([0, 1]), 2)
        layer = layers.CutMix(seed=2024)

        outputs = layer({"images": xs, "labels": ys})
        xs, ys = outputs["images"], outputs["labels"]

        # At least some pixels should be replaced in the CutMix operation
        self.assertTrue(tf.math.reduce_any(xs[0] == 1.0))
        self.assertTrue(tf.math.reduce_any(xs[0] == 2.0))
        self.assertTrue(tf.math.reduce_any(xs[1] == 1.0))
        self.assertTrue(tf.math.reduce_any(xs[1] == 2.0))
        # No labels should still be close to their original values
        self.assertNotAllClose(ys, 1.0)
        self.assertNotAllClose(ys, 0.0)

    def test_single_image_input(self):
        xs = tf.ones((4, 4, 3))
        ys = tf.one_hot(tf.constant([1]), 2)
        inputs = {"images": xs, "labels": ys}
        layer = layers.CutMix()

        with self.assertRaisesRegexp(
            ValueError, "CutMix received a single image to `call`"
        ):
            _ = layer(inputs)

    def test_missing_labels(self):
        xs = tf.ones((2, 4, 4, 3))
        inputs = {"images": xs}
        layer = layers.CutMix()

        with self.assertRaisesRegexp(ValueError, "CutMix expects `labels`"):
            _ = layer(inputs)

    def test_image_input(self):
        xs = tf.ones((2, 4, 4, 3))
        layer = layers.CutMix()

        with self.assertRaisesRegexp(
            ValueError,
            "CutMix expects `labels` or `segmentation_masks` to be present",
        ):
            _ = layer(xs)

    def test_cut_mix_call_segmentation_masks(self):
        xs = tf.cast(
            tf.stack(
                [2 * tf.ones((40, 40, 3)), tf.ones((40, 40, 3))],
                axis=0,
            ),
            tf.float32,
        )
        masks = tf.cast(
            tf.stack(
                [2 * tf.ones((40, 40, 1)), tf.ones((40, 40, 1))],
                axis=0,
            ),
            tf.float32,
        )
        ys = tf.one_hot(tf.constant([0, 1]), 2)
        layer = layers.CutMix(seed=2024)

        outputs = layer(
            {"images": xs, "labels": ys, "segmentation_masks": masks}
        )
        xs, ys = outputs["images"], outputs["labels"]
        masks = outputs["segmentation_masks"]

        # At least some pixels should be replaced in the CutMix operation
        self.assertTrue(tf.math.reduce_any(masks[0] == 1.0))
        self.assertTrue(tf.math.reduce_any(masks[0] == 2.0))
        self.assertTrue(tf.math.reduce_any(masks[1] == 1.0))
        self.assertTrue(tf.math.reduce_any(masks[1] == 2.0))



import numpy as np
import tensorflow as tf
from absl.testing import parameterized

from keras_aug import layers
from keras_aug.datapoints import bounding_box


class PadIfNeededTest(tf.test.TestCase, parameterized.TestCase):
    height, width = 32, 32
    regular_args = {
        "min_height": 32,
        "min_width": 32,
        "pad_height_divisor": None,
        "pad_width_divisor": None,
        "position": "center",
        "padding_value": 0,
        "bounding_box_format": "rel_xyxy",
    }

    def test_augments_image(self):
        input_image_shape = (4, 50, 60, 3)
        image = tf.random.uniform(shape=input_image_shape)
        layer = layers.PadIfNeeded(**self.regular_args, seed=2023)
        input_image_resized = tf.image.resize(image, [self.height, self.width])

        output = layer(image)

        self.assertNotAllClose(output, input_image_resized)

    def test_grayscale(self):
        input_image_shape = (4, self.height, self.width, 1)
        image = tf.random.uniform(shape=input_image_shape)
        layer = layers.PadIfNeeded(**self.regular_args, seed=2023)

        output = layer(image)

        self.assertAllEqual(output.shape, (4, self.height, self.width, 1))

    def test_augment_sparse_segmentation_mask(self):
        num_classes = 8
        input_image_shape = (1, 50, 50, 3)
        mask_shape = (1, 50, 50, 1)
        image = tf.random.uniform(shape=input_image_shape, seed=2023)
        mask = np.random.randint(2, size=mask_shape) * (num_classes - 1)
        inputs = {"images": image, "segmentation_masks": mask}

        # Crop-only to exactly 1/2 of the size
        args = self.regular_args.copy()
        args.update({"min_height": 100, "min_width": 100})
        layer = layers.PadIfNeeded(**args, seed=2023)

        output = layer(inputs)

        self.assertAllClose(
            output["segmentation_masks"][:, 25:-25, 25:-25, :], mask
        )

        # Crop to an arbitrary size and make sure we don't do bad interpolation
        args = self.regular_args.copy()
        args.update({"min_height": 123, "min_width": 123})
        layer = layers.PadIfNeeded(**args, seed=2023)

        output = layer(inputs)

        self.assertAllInSet(output["segmentation_masks"], [0, num_classes - 1])

    def test_augment_bounding_box_single(self):
        image = tf.zeros([20, 20, 3])
        boxes = {
            "boxes": tf.convert_to_tensor([[0, 0, 1, 1]]),
            "classes": tf.convert_to_tensor([0]),
        }
        input = {"images": image, "bounding_boxes": boxes}
        args = self.regular_args.copy()
        args.update({"min_height": 30, "min_width": 30})
        layer = layers.PadIfNeeded(**args, seed=2023)
        expected_output = {
            "boxes": tf.convert_to_tensor([[0, 0, 1, 1]], dtype=tf.float32),
            "classes": tf.convert_to_tensor([0], dtype=tf.float32),
        }

        output = layer(input)
        output["bounding_boxes"] = bounding_box.to_dense(
            output["bounding_boxes"]
        )

        self.assertNotAllClose(
            expected_output["boxes"], output["bounding_boxes"]["boxes"]
        )
        self.assertAllClose(
            expected_output["classes"], output["bounding_boxes"]["classes"]
        )

    def test_augment_boxes_batched_input(self):
        image = tf.zeros([20, 20, 3])
        boxes = {
            "boxes": tf.convert_to_tensor(
                [
                    [[0, 0, 1, 1], [0, 0, 1, 1]],
                    [[0, 0, 1, 1], [0, 0, 1, 1]],
                ]
            ),
            "classes": tf.convert_to_tensor([[0, 0], [0, 0]]),
        }
        input = {"images": [image, image], "bounding_boxes": boxes}
        args = self.regular_args.copy()
        args.update({"min_height": 30, "min_width": 30})
        layer = layers.PadIfNeeded(**args, seed=2023)
        expected_output = {
            "boxes": tf.convert_to_tensor(
                [
                    [[0, 0, 1, 1], [0, 0, 1, 1]],
                    [[0, 0, 1, 1], [0, 0, 1, 1]],
                ]
            ),
            "classes": tf.convert_to_tensor([[0, 0], [0, 0]]),
        }

        output = layer(input)
        output["bounding_boxes"] = bounding_box.to_dense(
            output["bounding_boxes"]
        )

        self.assertNotAllClose(
            expected_output["boxes"], output["bounding_boxes"]["boxes"]
        )
        self.assertAllClose(
            expected_output["classes"], output["bounding_boxes"]["classes"]
        )

    def test_augment_boxes_ragged(self):
        image = tf.zeros([2, 20, 20, 3])
        boxes = {
            "boxes": tf.ragged.constant(
                [[[0, 0, 1, 1], [0, 0, 1, 1]], [[0, 0, 1, 1]]], dtype=tf.float32
            ),
            "classes": tf.ragged.constant([[0, 0], [0]]),
        }
        input = {"images": image, "bounding_boxes": boxes}
        args = self.regular_args.copy()
        args.update({"min_height": 30, "min_width": 30})
        layer = layers.PadIfNeeded(**args, seed=2023)
        # the result boxes will still have the entire image in them
        expected_output = {
            "boxes": tf.ragged.constant(
                [[[0, 0, 1, 1], [0, 0, 1, 1]], [[0, 0, 1, 1]]], dtype=tf.float32
            ),
            "classes": tf.ragged.constant([[0, 0], [0]]),
        }
        expected_output = bounding_box.to_dense(expected_output)

        output = layer(input)
        output["bounding_boxes"] = bounding_box.to_dense(
            output["bounding_boxes"]
        )

        self.assertNotAllClose(
            expected_output["boxes"], output["bounding_boxes"]["boxes"]
        )
        self.assertAllClose(
            expected_output["classes"], output["bounding_boxes"]["classes"]
        )

    def test_padding_center(self):
        inputs = tf.ones((1, 4, 8, 3))
        args = self.regular_args.copy()
        args.update({"min_height": 8, "min_width": 8, "position": "center"})
        layer = layers.PadIfNeeded(**args)

        outputs = layer(inputs)

        self.assertEqual(outputs.shape, (1, 8, 8, 3))
        self.assertEqual(tf.reduce_mean(outputs[:, 0:2, :, :]), 0.0)
        self.assertEqual(tf.reduce_mean(outputs[:, -2:, :, :]), 0.0)

    def test_padding_top(self):
        inputs = tf.ones((1, 4, 8, 3))
        args = self.regular_args.copy()
        args.update({"min_height": 8, "min_width": 8, "position": "top_left"})
        layer = layers.PadIfNeeded(**args)

        outputs = layer(inputs)

        self.assertEqual(outputs.shape, (1, 8, 8, 3))
        self.assertNotEqual(tf.reduce_mean(outputs[:, 0:2, :, :]), 0.0)
        self.assertEqual(tf.reduce_mean(outputs[:, -4:, :, :]), 0.0)

    def test_padding_bottom(self):
        inputs = tf.ones((1, 4, 8, 3))
        args = self.regular_args.copy()
        args.update(
            {"min_height": 8, "min_width": 8, "position": "bottom_left"}
        )
        layer = layers.PadIfNeeded(**args)

        outputs = layer(inputs)

        self.assertEqual(outputs.shape, (1, 8, 8, 3))
        self.assertEqual(tf.reduce_mean(outputs[:, 0:4, :, :]), 0.0)
        self.assertNotEqual(tf.reduce_mean(outputs[:, -2:, :, :]), 0.0)

    def test_padding_left(self):
        inputs = tf.ones((1, 8, 4, 3))
        args = self.regular_args.copy()
        args.update({"min_height": 8, "min_width": 8, "position": "top_left"})
        layer = layers.PadIfNeeded(**args)

        outputs = layer(inputs)

        self.assertEqual(outputs.shape, (1, 8, 8, 3))
        self.assertNotEqual(tf.reduce_mean(outputs[:, :, 0:2, :]), 0.0)
        self.assertEqual(tf.reduce_mean(outputs[:, :, -2:, :]), 0.0)

    def test_padding_right(self):
        inputs = tf.ones((1, 8, 4, 3))
        args = self.regular_args.copy()
        args.update({"min_height": 8, "min_width": 8, "position": "top_right"})
        layer = layers.PadIfNeeded(**args)

        outputs = layer(inputs)

        self.assertEqual(outputs.shape, (1, 8, 8, 3))
        self.assertEqual(tf.reduce_mean(outputs[:, :, 0:4, :]), 0.0)
        self.assertNotEqual(tf.reduce_mean(outputs[:, :, -2:, :]), 0.0)

    def test_dense_segmentation_masks(self):
        images = tf.random.uniform((2, 10, 10, 3))
        segmentation_masks = tf.random.uniform(
            (2, 10, 10, 1), minval=0, maxval=10, dtype=tf.int32
        )
        args = self.regular_args.copy()
        args.update(
            {"min_height": 16, "min_width": 16, "position": "top_right"}
        )
        layer = layers.PadIfNeeded(**args)

        result = layer(
            {"images": images, "segmentation_masks": segmentation_masks}
        )

        self.assertTrue(isinstance(result["segmentation_masks"], tf.Tensor))
        self.assertEqual(result["segmentation_masks"].shape[1:3], (16, 16))
        self.assertAllInSet(result["segmentation_masks"], tf.range(0, 10))

    def test_ragged_segmentation_masks(self):
        images = tf.ragged.stack(
            [
                tf.random.uniform((8, 8, 3), dtype=tf.float32),
                tf.random.uniform((16, 8, 3), dtype=tf.float32),
            ]
        )
        segmentation_masks = tf.ragged.stack(
            [
                tf.random.uniform((8, 8, 1), maxval=10, dtype=tf.int32),
                tf.random.uniform((16, 8, 1), maxval=10, dtype=tf.int32),
            ]
        )
        segmentation_masks = tf.cast(segmentation_masks, dtype=tf.float32)
        args = self.regular_args.copy()
        args.update(
            {"min_height": 16, "min_width": 16, "position": "top_right"}
        )
        layer = layers.PadIfNeeded(**args)

        result = layer(
            {"images": images, "segmentation_masks": segmentation_masks}
        )

        self.assertTrue(
            isinstance(result["segmentation_masks"], tf.RaggedTensor)
        )
        self.assertAllInSet(
            result["segmentation_masks"].to_tensor(), tf.range(0, 10)
        )



import pytest
import tensorflow as tf


def pytest_configure():
    physical_devices = tf.config.list_physical_devices("GPU")
    for device in physical_devices:
        tf.config.experimental.set_memory_growth(device, True)


def pytest_addoption(parser):
    parser.addoption(
        "--skip-large",
        action="store_true",
        default=False,
        help="skip large tests",
    )


def pytest_collection_modifyitems(config, items):
    skip_large_tests = config.getoption("--skip-large")
    skip_large = pytest.mark.skipif(
        skip_large_tests, reason="need no --skip-large option to run"
    )
    for item in items:
        if "large" in item.keywords:
            item.add_marker(skip_large)



import tensorflow as tf
from absl.testing import parameterized

from keras_aug import layers


class RandomJpegQualityTest(tf.test.TestCase, parameterized.TestCase):
    regular_args = {
        "value_range": (0, 255),
        "factor": (75, 100),  # 25%
    }

    def test_adjustment_for_non_rgb_value_range(self):
        image_shape = (4, 8, 8, 3)
        # Value range (0, 100)
        image = tf.random.uniform(shape=image_shape) * 100.0
        args = self.regular_args.copy()
        args.update({"value_range": (0, 100)})
        layer = layers.RandomJpegQuality(**args)

        output = layer(image)

        self.assertNotAllClose(image, output)

    def test_in_single_image(self):
        layer = layers.RandomJpegQuality(**self.regular_args)
        # RGB
        xs = tf.cast(
            tf.ones((512, 512, 3)) * 255.0,
            dtype=tf.float32,
        )

        xs = layer(xs)

        self.assertEqual(xs.shape, [512, 512, 3])

        # greyscale
        xs = tf.cast(
            tf.ones((512, 512, 1)) * 255.0,
            dtype=tf.float32,
        )

        xs = layer(xs)

        self.assertEqual(xs.shape, [512, 512, 1])



import keras
import tensorflow as tf

from keras_aug.core.factor_sampler.factor_sampler import FactorSampler
from keras_aug.core.keras_random_generator import KerasRandomGenerator


@keras.utils.register_keras_serializable(package="keras_aug")
class NormalFactorSampler(FactorSampler):
    """NormalFactorSampler samples factors from a normal distribution.

    Args:
        mean (float): The mean value for the distribution.
        stddev (float): The standard deviation of the distribution.
        min_value (float): The minimum value.
        max_value (float): The maximum value.
        seed (int|float, optional): The random seed. Defaults to ``None``.

    References:
        - `KerasCV <https://github.com/keras-team/keras-cv>`_
    """

    def __init__(self, mean, stddev, min_value, max_value, seed=None):
        self.mean = mean
        self.stddev = stddev
        self.min_value = min_value
        self.max_value = max_value
        self.seed = seed
        self.rng = KerasRandomGenerator(seed=seed, rng_type="stateless")

    def __call__(self, shape=(), dtype="float32"):
        return tf.clip_by_value(
            self.rng.random_normal(
                shape=shape,
                mean=self.mean,
                stddev=self.stddev,
                dtype=dtype,
            ),
            self.min_value,
            self.max_value,
        )

    def get_config(self):
        return {
            "mean": self.mean,
            "stddev": self.stddev,
            "min_value": self.min_value,
            "max_value": self.max_value,
            "seed": self.seed,
        }



import tensorflow as tf

from keras_aug import layers


class GrayscaleTest(tf.test.TestCase):
    def test_return_shapes(self):
        xs = tf.ones((2, 4, 8, 3))

        layer = layers.Grayscale(output_channels=1)
        xs1 = layer(xs)

        layer = layers.Grayscale(output_channels=3)
        xs2 = layer(xs)

        self.assertEqual(xs1.shape, [2, 4, 8, 1])
        self.assertEqual(xs2.shape, [2, 4, 8, 3])

    def test_non_square_image(self):
        xs = tf.cast(
            tf.stack([2 * tf.ones((4, 8, 3)), tf.ones((4, 8, 3))], axis=0),
            tf.float32,
        )

        layer = layers.Grayscale(output_channels=1)
        xs1 = layer(xs)

        layer = layers.Grayscale(output_channels=3)
        xs2 = layer(xs)

        self.assertEqual(xs1.shape, [2, 4, 8, 1])
        self.assertEqual(xs2.shape, [2, 4, 8, 3])

    def test_in_single_image(self):
        xs = tf.cast(
            tf.ones((4, 8, 3)),
            dtype=tf.float32,
        )

        layer = layers.Grayscale(output_channels=1)
        xs1 = layer(xs)

        layer = layers.Grayscale(output_channels=3)
        xs2 = layer(xs)

        self.assertEqual(xs1.shape, [4, 8, 1])
        self.assertEqual(xs2.shape, [4, 8, 3])



import tensorflow as tf
from absl.testing import parameterized

from keras_aug import layers


class RandomBlurTest(tf.test.TestCase, parameterized.TestCase):
    regular_args = {
        "factor": (3, 7),
    }
    no_aug_args = {
        "factor": (1, 1),
    }

    def test_no_adjustment(self):
        image_shape = (4, 8, 8, 3)
        image = tf.random.uniform(shape=image_shape) * 255.0

        layer = layers.RandomBlur(**self.no_aug_args)
        output = layer(image)
        self.assertAllClose(image, output, atol=1e-5, rtol=1e-5)

    def test_adjustment_for_non_rgb_value_range(self):
        image_shape = (4, 8, 8, 3)
        # Value range (0, 100)
        image = tf.random.uniform(shape=image_shape) * 100.0

        layer = layers.RandomBlur(**self.no_aug_args)
        output = layer(image)
        self.assertAllClose(image, output, rtol=1e-5, atol=1e-5)

        layer = layers.RandomBlur(**self.regular_args)
        output = layer(image)
        self.assertNotAllClose(image, output)



import tensorflow as tf

import keras_aug


class UniformFactorSamplerTest(tf.test.TestCase):
    def test_sample(self):
        factor = keras_aug.UniformFactorSampler(0.3, 0.6)

        self.assertTrue(0.3 <= factor() <= 0.6)

    def test_config(self):
        factor = keras_aug.UniformFactorSampler(0.3, 0.6)

        config = factor.get_config()

        self.assertEqual(config["lower"], 0.3)
        self.assertEqual(config["upper"], 0.6)

    def test_from_config(self):
        factor = keras_aug.UniformFactorSampler(0.3, 0.6)
        config = factor.get_config()

        factor2 = keras_aug.UniformFactorSampler.from_config(config)
        config = factor2.get_config()

        self.assertEqual(config["lower"], 0.3)
        self.assertEqual(config["upper"], 0.6)
